""" MultiQC module to parse output from ngsderive """


import csv
import io
import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, linegraph, table

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """
    ngsderive module class, parses stderr logs.
    """

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="ngsderive",
            anchor="ngsderive",
            href="https://github.com/stjudecloud/ngsderive",
            info="attempts to predict library information from next-generation sequencing data.",
            # Can't find a DOI // doi=
        )

        self.strandedness = {}
        self.instrument = {}
        self.readlen = {}
        self.encoding = {}
        self.junctions = {}

        # parse ngsderive summary file
        expected_header_count_strandedness = 5
        expected_header_count_instrument = 4
        expected_header_count_readlen = 4
        expected_header_count_encoding = 3
        expected_header_count_junctions = 9

        for f in self.find_log_files("ngsderive/strandedness"):
            self.parse(
                self.strandedness,
                f,
                "strandedness",
                expected_header_count_strandedness,
            )

        for f in self.find_log_files("ngsderive/instrument"):
            self.parse(
                self.instrument,
                f,
                "instrument",
                expected_header_count_instrument,
            )

        for f in self.find_log_files("ngsderive/readlen"):
            self.parse(
                self.readlen,
                f,
                "readlen",
                expected_header_count_readlen,
            )

        for f in self.find_log_files("ngsderive/encoding"):
            self.parse(
                self.encoding,
                f,
                "encoding",
                expected_header_count_encoding,
            )

        for f in self.find_log_files("ngsderive/junction_annotation"):
            self.parse(
                self.junctions,
                f,
                "junctions",
                expected_header_count_junctions,
            )

        self.strandedness = self.ignore_samples(self.strandedness)
        self.instrument = self.ignore_samples(self.instrument)
        self.readlen = self.ignore_samples(self.readlen)
        self.encoding = self.ignore_samples(self.encoding)
        self.junctions = self.ignore_samples(self.junctions)

        num_results_found = max(
            [len(d) for d in [self.strandedness, self.instrument, self.readlen, self.encoding, self.junctions]]
        )
        if num_results_found == 0:
            raise ModuleNoSamplesFound
        log.info(f"Found {num_results_found} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        if self.strandedness:
            self.add_strandedness_data()

        if self.instrument:
            self.add_instrument_data()

        if self.readlen:
            self.add_readlen_data()

        if self.encoding:
            self.add_encoding_data()

        if self.junctions:
            self.add_junctions_data()

    @staticmethod
    def probe_file_for_dictreader_kwargs(f, expected_header_count):
        """In short, this function was created to figure out which
        kwargs need to be passed to csv.DictReader. First,
        it tries to extract commonly known delimiters without
        doing an expensive scanning operation. If that
        doesn't work, we try sniffing for a dialect."""

        header = f.readline()
        f.seek(0)

        for delim in ["\t", ",", " ", "|"]:
            if delim in header and len(header.split(delim)) == expected_header_count:
                return {"delimiter": delim}

        log.warning("Could not easily detect delimiter for file. Trying csv.Sniffer()")
        dialect = csv.Sniffer().sniff(f.read(4096))  # 4096 is the buffer size
        f.seek(0)
        return {"dialect": dialect}

    def parse(self, sample_dict, found_file, subcommand, expected_header_count):
        kwargs = self.probe_file_for_dictreader_kwargs(io.StringIO(found_file["f"]), expected_header_count)
        for row in csv.DictReader(io.StringIO(found_file["f"]), **kwargs):
            if not row.get("File"):
                continue
            sample_name = self.clean_s_name(row.get("File"), found_file)
            if sample_name in sample_dict:
                log.debug(f"Duplicate sample name found! Overwriting: {sample_name}")

            sample_dict[sample_name] = row
            self.add_data_source(f=found_file, s_name=sample_name)

    def add_strandedness_data(self):
        # Write data to file
        self.write_data_file(self.strandedness, "ngsderive_strandedness")

        data = {}
        for sample, strandedness in self.strandedness.items():
            data[sample] = {
                "predicted": strandedness.get("Predicted"),
                "forward": round(float(strandedness.get("ForwardPct")) * 100.0, 2),
                "reverse": round(float(strandedness.get("ReversePct")) * 100.0, 2),
            }

        bardata = {}
        sorted_data = sorted(data.items(), key=lambda x: x[1].get("forward"))
        for k, v in sorted_data:
            bardata[k] = v

        headers = {
            "predicted": {
                "title": "Strandedness",
                "description": "Predicted strandedness from ngsderive",
                "scale": False,
            }
        }
        self.general_stats_addcols(data, headers)

        # Config for the plot
        pconfig = {
            "id": "ngsderive_strandedness_plot",
            "title": "ngsderive: Strandedness",
            "ylab": "% Read Evidence",
            "ymin": 0,
            "ymax": 100,
            "tt_percentages": True,
            "ylab_format": "{value}%",
            "cpswitch": False,
        }

        self.add_section(
            name="Strandedness",
            anchor="ngsderive-strandedness",
            description="""Predicted strandedness provided by ngsderive. For more information, please see
            [the documentation](https://stjudecloud.github.io/ngsderive/subcommands/strandedness/).""",
            plot=bargraph.plot(bardata, ["forward", "reverse"], pconfig),
        )

    def add_instrument_data(self):
        # Write data to file
        self.write_data_file(self.instrument, "ngsderive_instrument")

        bgcols = {
            "low confidence": "#f8d7da",
            "medium confidence": "#fff3cd",
            "high confidence": "#d1e7dd",
        }
        cond_formatting_rules = {
            "pass": [{"s_eq": "high confidence"}],
            "warn": [{"s_eq": "medium confidence"}],
            "fail": [{"s_eq": "low confidence"}],
        }

        general_data = {}
        for sample, instrument_data in self.instrument.items():
            general_data[sample] = {
                "instrument": " / ".join(sorted(instrument_data.get("Instrument").split(" or "))),
                "confidence": instrument_data.get("Confidence"),
            }

        general_headers = {
            "instrument": {
                "title": "Predicted Instrument",
                "description": "Predicted instrument from ngsderive",
            },
            "confidence": {
                "title": "Instrument: Confidence",
                "description": "Level of confidence (low, medium, high) that the predicted instrument is correct.",
                "bgcols": bgcols,
                "cond_formatting_rules": cond_formatting_rules,
                "hidden": True,
            },
        }
        self.general_stats_addcols(general_data, general_headers)

        instruments = set()

        for d in general_data.values():
            instruments.update(d.get("instrument").split(" / "))

        # move multiple instruments to the end if it exists
        instruments = sorted(instruments)
        if "multiple instruments" in instruments:
            instruments.remove("multiple instruments")
            instruments.append("multiple instruments")

        headers = {}
        for instrument in instruments:
            headers[instrument] = {
                "title": instrument,
                "description": f"Predicted {instrument} from ngsderive",
                "bgcols": bgcols,
                "cond_formatting_rules": cond_formatting_rules,
            }
        headers["basis"] = {
            "title": "Instrument: Basis",
            "description": "Basis upon which the prediction was made.",
        }

        table_data = {}
        for sample, instrument_data in self.instrument.items():
            table_data[sample] = {}
            for instrument in instrument_data.get("Instrument").split(" or "):
                table_data[sample][instrument] = instrument_data.get("Confidence")
            table_data[sample]["basis"] = instrument_data.get("Basis")

        # Config for the plot
        config = {
            "id": "ngsderive_instruments_plot",
            "title": "ngsderive: Instruments",
            "scale": False,
        }

        self.add_section(
            name="Instrument",
            anchor="ngsderive-instrument",
            description="""Predicted instrument provided by ngsderive. For more information, please see
            [the documentation](https://stjudecloud.github.io/ngsderive/subcommands/instrument/).""",
            plot=table.plot(table_data, headers, config),
        )

    def add_readlen_data(self):
        # Write data to file
        self.write_data_file(self.readlen, "ngsderive_readlen")

        data = {}
        for sample, readlen in self.readlen.items():
            data[sample] = {
                "evidence": readlen.get("Evidence"),
                "majoritypctdetected": round(float(readlen.get("MajorityPctDetected")) * 100.0, 2),
                "consensusreadlength": int(readlen.get("ConsensusReadLength")),
            }

        headers = {
            "consensusreadlength": {
                "title": "Read Length (bp)",
                "description": "Predicted read length from ngsderive.",
                "format": "{:,d}",
            },
            "majoritypctdetected": {
                "title": "Read Length: % Supporting",
                "description": "Percentage of reads which were measured at the predicted read length.",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "hidden": True,
            },
        }
        self.general_stats_addcols(data, headers)

        linedata = [{}, {}]

        for sample, d in data.items():
            # Build dict of count data
            count_data = {}
            for parts in d.get("evidence").split(";"):
                (k, v) = parts.split("=")
                count_data[int(k)] = int(v)
            linedata[1][sample] = count_data

            # Build dict of percentage data
            total_reads = sum(count_data.values())
            linedata[0][sample] = {readlen: (count / total_reads) * 100.0 for readlen, count in count_data.items()}

        # Config for the plot
        pconfig = {
            "id": "ngsderive_readlen_plot",
            "title": "ngsderive: Read Length",
            "xlab": "Read Length",
            "ylab": "% Evidence for Read Length",
            "data_labels": [
                {"name": "Percentages", "ylab": "% Evidence for Read Length"},
                {"name": "Counts", "ylab": "Number of reads"},
            ],
        }

        self.add_section(
            name="Read length",
            anchor="ngsderive-readlen",
            description="""Predicted read length provided by ngsderive. For more information, please see
            [the documentation](https://stjudecloud.github.io/ngsderive/subcommands/readlen/).""",
            plot=linegraph.plot(linedata, pconfig),
        )

    def add_encoding_data(self):
        # Write data to file
        self.write_data_file(self.encoding, "ngsderive_encoding")

        general_data = {}
        for sample, encoding_data in self.encoding.items():
            general_data[sample] = {
                "probable_encoding": encoding_data.get("ProbableEncoding"),
                "evidence": encoding_data.get("Evidence"),
            }

        general_headers = {
            "probable_encoding": {
                "title": "Probable Encoding",
                "description": "Predicted PHRED score encoding from ngsderive",
            },
            "evidence": {
                "title": "Encoding: Evidence",
                "description": "Observed ASCII value ranges in PHRED score encoding",
                "hidden": True,
            },
        }
        self.general_stats_addcols(general_data, general_headers)

    def add_junctions_data(self):
        # Write data to file
        self.write_data_file(self.junctions, "ngsderive_junctions")

        data = {}
        for sample, junctions_data in self.junctions.items():
            data[sample] = {
                "total_junctions": junctions_data.get("total_junctions"),
                "known_junctions": junctions_data.get("known_junctions"),
                "partial_novel_junctions": junctions_data.get("partial_novel_junctions"),
                "novel_junctions": junctions_data.get("complete_novel_junctions"),
                "total_splice_events": junctions_data.get("total_splice_events"),
                "known_spliced_reads": junctions_data.get("known_spliced_reads"),
                "partial_novel_spliced_reads": junctions_data.get("partial_novel_spliced_reads"),
                "novel_spliced_reads": junctions_data.get("complete_novel_spliced_reads"),
            }

        headers = {
            "total_junctions": {
                "title": "Total junctions",
                "description": "Total number of junctions found by ngsderive",
                "hidden": True,
                "format": "{:,d}",
            },
            "known_junctions": {
                "title": "Known junctions",
                "description": "Number of annotated junctions found by ngsderive",
                "format": "{:,d}",
            },
            "partial_novel_junctions": {
                "title": "Partially novel junctions",
                "description": "Number of partially annotated junctions found by ngsderive",
                "format": "{:,d}",
            },
            "novel_junctions": {
                "title": "Novel junctions",
                "description": "Number of completely novel junctions found by ngsderive",
                "format": "{:,d}",
            },
            "total_splice_events": {
                "title": "Total splice events",
                "description": "Total number of spliced reads found by ngsderive",
                "hidden": True,
                "format": "{:,d}",
            },
            "known_spliced_reads": {
                "title": "Annotated spliced reads",
                "description": "Number of annotated spliced reads found by ngsderive",
                "hidden": True,
                "format": "{:,d}",
            },
            "partial_novel_spliced_reads": {
                "title": "Partially annotated spliced reads",
                "description": "Number of partially annotated spliced reads found by ngsderive",
                "hidden": True,
                "format": "{:,d}",
            },
            "novel_spliced_reads": {
                "title": "Novel spliced reads",
                "description": "Number of completely un-annotated spliced reads found by ngsderive",
                "hidden": True,
                "format": "{:,d}",
            },
        }
        self.general_stats_addcols(data, headers)

        bardata = {}
        sorted_junction_data = sorted(data.items(), key=lambda x: int(x[1].get("total_junctions")), reverse=True)
        for k, v in sorted_junction_data:
            bardata[k] = {
                "known_junctions": v["known_junctions"],
                "partial_novel_junctions": v["partial_novel_junctions"],
                "novel_junctions": v["novel_junctions"],
                "known_spliced_reads": v["known_spliced_reads"],
                "partial_novel_spliced_reads": v["partial_novel_spliced_reads"],
                "novel_spliced_reads": v["novel_spliced_reads"],
            }

        # Config for the plot
        pconfig = {
            "id": "ngsderive_junctions_plot",
            "title": "ngsderive: Junction Annotation",
            "cpswitch_counts_label": "Number",
            "yDecimals": False,
            "ylab": "Number of junctions",
            "data_labels": [
                {"name": "Junctions", "ylab": "Number of junctions"},
                {"name": "Spliced Reads", "ylab": "Number of spliced reads"},
            ],
        }

        cats = [
            {
                "known_junctions": {"name": "Known junctions"},
                "partial_novel_junctions": {"name": "Partially novel junctions"},
                "novel_junctions": {"name": "Novel junctions"},
            },
            {
                "known_spliced_reads": {"name": "Annotated spliced reads"},
                "partial_novel_spliced_reads": {"name": "Partially annotated spliced reads"},
                "novel_spliced_reads": {"name": "Novel spliced reads"},
            },
        ]
        self.add_section(
            name="Junction Annotations",
            anchor="ngsderive-junctions",
            description="""Junction annotations provided by ngsderive. For more information, please see
            [the documentation](https://stjudecloud.github.io/ngsderive/subcommands/junction_annotation/).""",
            plot=bargraph.plot([bardata, bardata], cats, pconfig),
        )



""" MultiQC submodule to parse output from RSeQC junction_annotation.py
http://rseqc.sourceforge.net/#junction-annotation-py """

import logging
import re

from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(self):
    """Find RSeQC junction_annotation reports and parse their data"""

    # Set up vars
    self.junction_annotation_data = dict()
    regexes = {
        "total_splicing_events": r"^Total splicing  Events:\s*(\d+)$",
        "known_splicing_events": r"^Known Splicing Events:\s*(\d+)$",
        "partial_novel_splicing_events": r"^Partial Novel Splicing Events:\s*(\d+)$",
        "novel_splicing_events": r"^Novel Splicing Events:\s*(\d+)$",
        "total_splicing_junctions": r"^Total splicing  Junctions:\s*(\d+)$",
        "known_splicing_junctions": r"^Known Splicing Junctions:\s*(\d+)$",
        "partial_novel_splicing_junctions": r"^Partial Novel Splicing Junctions:\s*(\d+)$",
        "novel_splicing_junctions": r"^Novel Splicing Junctions:\s*(\d+)$",
    }

    # Go through files and parse data using regexes
    for f in self.find_log_files("rseqc/junction_annotation"):
        d = dict()
        for k, r in regexes.items():
            r_search = re.search(r, f["f"], re.MULTILINE)
            if r_search:
                d[k] = int(r_search.group(1))

        # Calculate some percentages
        if "total_splicing_events" in d:
            t = float(d["total_splicing_events"])
            if "known_splicing_events" in d:
                d["known_splicing_events_pct"] = (float(d["known_splicing_events"]) / t) * 100.0
            if "partial_novel_splicing_events" in d:
                d["partial_novel_splicing_events_pct"] = (float(d["partial_novel_splicing_events"]) / t) * 100.0
            if "novel_splicing_events" in d:
                d["novel_splicing_events_pct"] = (float(d["novel_splicing_events"]) / t) * 100.0
        if "total_splicing_junctions" in d:
            t = float(d["total_splicing_junctions"])
            if "known_splicing_junctions" in d:
                d["known_splicing_junctions_pct"] = (float(d["known_splicing_junctions"]) / t) * 100.0
            if "partial_novel_splicing_junctions" in d:
                d["partial_novel_splicing_junctions_pct"] = (float(d["partial_novel_splicing_junctions"]) / t) * 100.0
            if "novel_splicing_junctions" in d:
                d["novel_splicing_junctions_pct"] = (float(d["novel_splicing_junctions"]) / t) * 100.0

        if len(d) > 0:
            if f["s_name"] in self.junction_annotation_data:
                log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
            self.add_data_source(f, section="junction_annotation")
            self.junction_annotation_data[f["s_name"]] = d

    # Filter to strip out ignored sample names
    self.junction_annotation_data = self.ignore_samples(self.junction_annotation_data)

    if len(self.junction_annotation_data) == 0:
        return 0

    # Write to file
    self.write_data_file(self.junction_annotation_data, "multiqc_rseqc_junction_annotation")

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    self.add_software_version(None)

    # Plot junction annotations
    keys = [
        {
            "known_splicing_junctions": {"name": "Known Splicing Junctions"},
            "partial_novel_splicing_junctions": {"name": "Partial Novel Splicing Junctions"},
            "novel_splicing_junctions": {"name": "Novel Splicing Junctions"},
        },
        {
            "known_splicing_events": {"name": "Known Splicing Events"},
            "partial_novel_splicing_events": {"name": "Partial Novel Splicing Events"},
            "novel_splicing_events": {"name": "Novel Splicing Events"},
        },
    ]
    pconfig = {
        "id": "rseqc_junction_annotation_junctions_plot",
        "title": "RSeQC: Splicing Junctions",
        "ylab": "% Junctions",
        "cpswitch_c_active": False,
        "data_labels": ["Junctions", "Events"],
    }
    self.add_section(
        name="Junction Annotation",
        anchor="rseqc_junction_annotation",
        description='<a href="http://rseqc.sourceforge.net/#junction-annotation-py" target="_blank">Junction annotation</a>'
        " compares detected splice junctions to"
        " a reference gene model. An RNA read can be spliced 2"
        " or more times, each time is called a splicing event.",
        plot=bargraph.plot([self.junction_annotation_data, self.junction_annotation_data], keys, pconfig),
    )

    # Return number of samples found
    return len(self.junction_annotation_data)



""" MultiQC submodule to parse output from deepTools plotCorrelation """

import logging

from multiqc.plots import heatmap

# Initialise the logger
log = logging.getLogger(__name__)


class plotCorrelationMixin:
    def parse_plotCorrelation(self):
        """Find plotCorrelation output"""
        self.deeptools_plotCorrelationData = dict()
        for f in self.find_log_files("deeptools/plotCorrelationData", filehandles=False):
            parsed_data = self.parsePlotCorrelationData(f)
            for sample, val_by_sample in parsed_data.items():
                if sample in self.deeptools_plotCorrelationData:
                    log.warning(f"Replacing duplicate sample {sample}.")
                self.deeptools_plotCorrelationData[sample] = val_by_sample
            if len(parsed_data) > 0:
                self.add_data_source(f, section="plotCorrelation")

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, f["s_name"])

        self.deeptools_plotCorrelationData = self.ignore_samples(self.deeptools_plotCorrelationData)
        for s_name, val_by_sample in self.deeptools_plotCorrelationData.items():
            self.deeptools_plotCorrelationData[s_name] = self.ignore_samples(val_by_sample)

        if len(self.deeptools_plotCorrelationData) > 0:
            # Data is in wrong format for writing to file
            # self.write_data_file(self.deeptools_plotCorrelationData, "deeptools_plot_corr")

            config = {
                "id": "deeptools_correlation_plot",
                "title": "deeptools: Correlation Plot",
            }
            samples = sorted(self.deeptools_plotCorrelationData.keys())
            rows = []
            for s1 in samples:
                row = []
                for s2 in samples:
                    try:
                        row.append(self.deeptools_plotCorrelationData[s1][s2])
                    except KeyError:
                        row.append(None)
                rows.append(row)
            if len(rows) == 0:
                log.debug("No valid data for correlation plot")
                return None

            self.add_section(
                name="Correlation heatmap",
                anchor="deeptools_correlation",
                description="Pairwise correlations of samples based on distribution of sequence reads",
                plot=heatmap.plot(rows, samples, samples, config),
            )

        return len(self.deeptools_plotCorrelationData)

    def parsePlotCorrelationData(self, f):
        d = dict()
        x_samples = None
        for line in f["f"].splitlines():
            cols = line.split("\t")
            if cols[0] == "#plotCorrelation --outFileCorMatrix":
                continue
            elif cols[0] == "":
                x_samples = [self.clean_s_name(s.strip("'")) for s in cols[1 : len(cols)]]
            else:
                y_sample = str(cols[0]).strip("'")
                y_sample = self.clean_s_name(y_sample, f)
                d[y_sample] = dict()
                for x_sample, col in zip(x_samples, cols[1 : len(cols)]):
                    d[y_sample][x_sample] = float(col)
        return d



""" Module to parse output from mOTUs """

import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph
from multiqc.utils import config

log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="motus",
            anchor="motus",
            href="https://motu-tool.org/",
            info="is a tool performing microbial profiling through marker gene (MG)-based operational taxonomic units (mOTUs).",
            doi="10.1038/s41467-019-08844-4",
        )

        # Define the main motus multiqc data object
        self.motus_data = dict()

        for f in self.find_log_files("motus", filehandles=True):
            self.parse_logs(f)

        self.motus_data = self.ignore_samples(self.motus_data)

        if len(self.motus_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.motus_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write data to file
        self.write_data_file(self.motus_data, "motus")

        self.motus_general_stats()
        self.motus_filtering_bargraph_plot()
        self.motus_mapping_bargraph_plot()
        self.motus_motus_bargraph_plot()

    def parse_logs(self, f):
        """Parses a motus stdout saved in a file"""
        # Assume take from the name of the file as not always reported in the log
        s_name = f["s_name"]

        if self.motus_data.get(s_name) is not None:
            log.warning(f"Duplicate sample name found based on filename! Overwriting: {s_name}")

        self.motus_data[s_name] = {}
        self.add_data_source(f, s_name=s_name)
        for line in f["f"]:
            # Search strings - keeping colon as to not pick up non-stats lines
            for cat in [
                "Total number of reads:",
                "Total number of inserts:",
                "Unique mappers:",
                "Multiple mappers:",
                "Ignored multiple mapper without unique hit:",
                "Number of ref-mOTUs:",
                "Number of meta-mOTUs:",
                "Number of ext-mOTUs:",
            ]:
                if cat in line:
                    self.motus_data[s_name][cat.replace(":", "")] = int(line.strip().split(":")[1].lstrip())
                elif "Number of reads after filtering:" in line:
                    self.motus_data[s_name]["Number of reads after filtering"] = int(
                        line.strip().split(":")[1].lstrip().split(" ")[0]
                    )
                    self.motus_data[s_name]["Percent reads after filtering"] = float(
                        line.strip().split(":")[1].lstrip().split(" ")[1].replace("(", "")
                    )
        self.motus_data[s_name]["Discarded reads"] = (
            self.motus_data[s_name]["Total number of reads"]
            - self.motus_data[s_name]["Number of reads after filtering"]
        )

    def motus_general_stats(self):
        """mOTUs read counts for general stats"""
        headers = {
            "Total number of reads": {
                "title": f"Total Input Reads ({config.read_count_prefix})",
                "description": f"Total number of input reads to mOTUs ({config.read_count_prefix})",
                "scale": "Greens",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
            },
            "Number of reads after filtering": {
                "title": f"Total Mapped Reads ({config.read_count_prefix})",
                "description": f"Total number of reads after mapping({config.read_count_prefix})",
                "scale": "Greens",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
            },
            "Total number of inserts": {
                "title": f"Total Mapped Inserts ({config.read_count_prefix})",
                "description": f"Total number of inserts mapped to a MGC ({config.read_count_prefix})",
                "scale": "Purples",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
            },
            "Unique mappers": {
                "title": f"Unique Mapped Inserts ({config.read_count_prefix})",
                "description": f"Total number of inserts mapped to a single MGC ({config.read_count_prefix})",
                "scale": "Greens",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
                "hidden": True,
            },
            "Multiple mappers": {
                "title": f" Multi-mapped Inserts ({config.read_count_prefix})",
                "description": f"Total number of inserts mapped to multiple MGCs ({config.read_count_prefix})",
                "scale": "Purples",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
                "hidden": True,
            },
            "Ignored multiple mapper without unique hit": {
                "title": f"Ignored Multi-mapped Inserts ({config.read_count_prefix})",
                "description": f"Total number of ignored multi-MGC mapped reads ({config.read_count_prefix})",
                "scale": "Greens",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
                "hidden": True,
            },
            "Number of ref-mOTUs": {
                "title": "Total ref-mOTUs",
                "description": "Total known species mOTUs found",
                "scale": "Purples",
                "shared_key": "mOTUs",
                "format": "{:,.0f}",
            },
            "Number of meta-mOTUs": {
                "title": "Total meta-mOTUs",
                "description": "Total number of unknown metagenome-derived mOTUs found",
                "scale": "Greens",
                "shared_key": "mOTUs",
                "format": "{:,.0f}",
            },
            "Number of ext-mOTUs": {
                "title": "Total ext-mOTUs",
                "description": "Total number of unknown MAG-derived mOTUs found",
                "scale": "Purples",
                "shared_key": "mOTUs",
                "format": "{:,.0f}",
            },
        }

        self.general_stats_addcols(self.motus_data, headers)

    def motus_filtering_bargraph_plot(self):
        """mOTUs read counts for general stats"""
        common = {
            "min": 0,
            "modify": lambda x: float(x) * config.read_count_multiplier,
            "suffix": f"{config.read_count_prefix} reads",
            "decimalPlaces": 0,
            "shared_key": "read_count",
        }
        cats = {
            "Number of reads after filtering": dict(common, **{"name": "Reads after mapping"}),
            "Discarded reads": dict(common, **{"name": "Unmapped reads"}),
        }

        self.add_section(
            name="mOTUs: Read filtering information",
            anchor="motus-filtering",
            description="Read filtering statistics (i.e. mapping of reads to the mOTUs marker database).",
            plot=bargraph.plot(
                self.motus_data,
                cats,
                {
                    "id": "motus-filtering-reads",
                    "title": "mOTUs: Read filtering information",
                    "ylab": "Reads",
                },
            ),
        )

    def motus_mapping_bargraph_plot(self):
        """mOTUs bar chart of insert types"""
        common = {
            "min": 0,
            "modify": lambda x: float(x) * config.read_count_multiplier,
            "suffix": f"{config.read_count_prefix} reads",
            "decimalPlaces": 0,
            "shared_key": "read_count",
        }
        cats = {
            "Unique mappers": dict(common, **{"name": "Unique mapped inserts", "color": "#3aba5e"}),
            "Multiple mappers": dict(common, **{"name": "Multiple mapped inserts", "color": "#ebbe59"}),
            "Ignored multiple mapper without unique hit": dict(
                common, **{"name": "Ignored multi-mapped inserts", "color": "#cf5565"}
            ),
        }

        self.add_section(
            name="mOTUs: Insert mapping information",
            anchor="motus-mapping",
            description="How inserts was classified after alignment to MGCs.",
            plot=bargraph.plot(
                self.motus_data,
                cats,
                {
                    "id": "motus-mapping-inserts",
                    "title": "mOTUs: Insert mapping information",
                    "ylab": "Inserts",
                },
            ),
        )

    def motus_motus_bargraph_plot(self):
        """mOTUs bar chart of mOTU types"""
        common = {
            "min": 0,
            "decimalPlaces": 0,
        }
        cats = {
            "Number of ref-mOTUs": dict(common, **{"name": "Known mOTUs"}),
            "Number of meta-mOTUs": dict(common, **{"name": "(Unknown) Metagenome mOTUs"}),
            "Number of ext-mOTUs": dict(common, **{"name": "(Unknown) MAG mOTUs"}),
        }

        self.add_section(
            name="mOTUs: mOTU identification information",
            anchor="motus-identification",
            description="Distribution of the types of mOTUs found.",
            plot=bargraph.plot(
                self.motus_data,
                cats,
                {
                    "id": "motus-identification-types",
                    "title": "mOTUs: mOTU identification information",
                    "ylab": "mOTUs",
                },
            ),
        )



#!/usr/bin/env python

"""
Code to initilise the MultiQC logging
"""

import logging
import os
import shutil
import sys
import tempfile

import coloredlogs

from multiqc.utils import config, util_functions

LEVELS = {0: "INFO", 1: "DEBUG"}
log_tmp_dir = None
log_tmp_fn = "/dev/null"


def init_log(logger, loglevel=0, no_ansi=False):
    """
    Initializes logging.
    Prints logs to console with level defined by loglevel
    Also prints verbose log to the multiqc data directory if available.
    (multiqc_data/multiqc.log)

    Args:
        loglevel (str): Determines the level of the log output.
    """
    # File for logging
    global log_tmp_dir, log_tmp_fn
    log_tmp_dir = tempfile.mkdtemp()
    log_tmp_fn = os.path.join(log_tmp_dir, "multiqc.log")

    # Logging templates
    debug_template = "[%(asctime)s] %(name)-50s [%(levelname)-7s]  %(message)s"
    info_template = "|%(module)18s | %(message)s"

    # Remove log handlers left from previous calls to multiqc.run
    while logger.handlers:
        logger.removeHandler(logger.handlers[0])

    # Base level setup
    logger.setLevel(getattr(logging, "DEBUG"))

    # Automatically set no_ansi if not a tty terminal
    if not no_ansi:
        if not sys.stderr.isatty() and not util_functions.force_term_colors():
            no_ansi = True

    # Set up the console logging stream
    console = logging.StreamHandler()
    console.setLevel(getattr(logging, loglevel))
    level_styles = coloredlogs.DEFAULT_LEVEL_STYLES
    level_styles["debug"] = {"faint": True}
    field_styles = coloredlogs.DEFAULT_FIELD_STYLES
    field_styles["module"] = {"color": "blue"}
    if loglevel == "DEBUG":
        if no_ansi:
            console.setFormatter(logging.Formatter(debug_template))
        else:
            console.setFormatter(
                coloredlogs.ColoredFormatter(fmt=debug_template, level_styles=level_styles, field_styles=field_styles)
            )
    else:
        if no_ansi:
            console.setFormatter(logging.Formatter(info_template))
        else:
            console.setFormatter(
                coloredlogs.ColoredFormatter(fmt=info_template, level_styles=level_styles, field_styles=field_styles)
            )
    logger.addHandler(console)

    # Now set up the file logging stream if we have a data directory
    file_handler = logging.FileHandler(log_tmp_fn, encoding="utf-8")
    file_handler.setLevel(getattr(logging, "DEBUG"))  # always DEBUG for the file
    file_handler.setFormatter(logging.Formatter(debug_template))
    logger.addHandler(file_handler)


def move_tmp_log(logger):
    """Move the temporary log file to the MultiQC data directory
    if it exists."""

    try:
        # https://stackoverflow.com/questions/15435652/python-does-not-release-filehandles-to-logfile
        logging.shutdown()
        shutil.copy(log_tmp_fn, os.path.join(config.data_dir, "multiqc.log"))
        os.remove(log_tmp_fn)
        util_functions.robust_rmtree(log_tmp_dir)
    except (AttributeError, TypeError, IOError):
        pass


def get_log_stream(logger):
    """
    Returns a stream to the root log file.
    If there is no logfile return the stderr log stream

    Returns:
        A stream to the root log file or stderr stream.
    """

    file_stream = None
    log_stream = None
    for handler in logger.handlers:
        if isinstance(handler, logging.FileHandler):
            file_stream = handler.stream
        else:
            log_stream = handler.stream

    if file_stream:
        return file_stream

    return log_stream



def clean_title_case(col_id):
    title = col_id.title() if col_id[0:1].islower() else col_id
    for str in ["Bc", "bc", "Umi", "Igk", "Igh", "Igl", "Vj", "q30"]:
        title = title.replace(str, str.upper())
    return title


def update_dict(table, headers, rows_list, col_map, colours, namespace):
    """update the data dict and headers dict"""

    for col_name, col_data in rows_list:
        if col_name in col_map:
            # Sanitize numeric data
            is_percentage = "%" in col_data
            col_data = col_data.replace(",", "").replace("%", "")

            # Convert to float when possible
            try:
                col_data = float(col_data)
            except ValueError:
                col_data = col_data

            col_id = col_map[col_name]
            table[col_id] = col_data
            headers[col_id] = {
                "rid": f"{namespace}_{col_id.replace(' ', '_').replace('/', '_')}".lower(),
                "title": clean_title_case(col_id),
                "description": col_name,
                "namespace": namespace,
                "scale": colours.get(col_id, "RdYlGn" if is_percentage else "GnBu"),
            }
            if is_percentage:
                headers[col_id].update(
                    {
                        "suffix": "%",
                        "max": 100,
                        "min": 0,
                    }
                )

            # Assign shared keys
            if col_id == "estimated cells":
                headers[col_id]["shared_key"] = "cell_count"
                headers[col_id]["format"] = "{:,.0f}"
            if col_id.endswith("reads/cell"):
                headers[col_id]["shared_key"] = "reads_per_cell"
                headers[col_id]["format"] = "{:,.0f}"

    return table, headers


def set_hidden_cols(headers, col_names):
    """Set the hidden columns"""

    for col_name in col_names:
        try:
            headers[col_name]["hidden"] = True
        except KeyError:
            pass

    return headers


def parse_bcknee_data(data, s_name, max_idx=1000):
    """parse data for bc knee plot from cellranger dict"""

    value_dict = dict()
    for idx, data_series in enumerate(data):
        if idx > max_idx:
            break
        if len(data_series["x"]) == 0:
            continue
        id = f"{s_name}_{data_series['name']}"
        if id not in value_dict.keys():
            value_dict[id] = dict()
        value_dict[id].update(transform_data(data_series))

    return value_dict


def transform_data(data):
    """Transform x:list,y:list data to a dict of x_val:y_val"""

    value_dict = dict()
    for idx, row in enumerate(data["x"]):
        if row > 0 and data["y"][idx] > 0:
            value_dict[row] = data["y"][idx]

    return value_dict



""" MultiQC module to parse output from Salmon """

import json
import logging
import os

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import linegraph
from multiqc.utils import config

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Salmon",
            anchor="salmon",
            href="https://combine-lab.github.io/salmon/",
            info="is a tool for quantifying the expression of transcripts using RNA-seq data.",
            doi="10.1038/nmeth.4197",
        )

        # Parse meta information. JSON win!
        self.salmon_meta = dict()
        for f in self.find_log_files("salmon/meta"):
            # Get the s_name from the parent directory
            if os.path.basename(f["root"]) in ["aux_info", "aux"]:
                s_name = os.path.basename(os.path.dirname(f["root"]))
                s_name = self.clean_s_name(s_name, f)
                self.salmon_meta[s_name] = json.loads(f["f"])
                self.add_software_version(self.salmon_meta[s_name]["salmon_version"], s_name)

        # Parse Fragment Length Distribution logs
        self.salmon_fld = dict()
        for f in self.find_log_files("salmon/fld"):
            # Get the s_name from the parent directory
            if os.path.basename(f["root"]) == "libParams":
                s_name = os.path.basename(os.path.dirname(f["root"]))
                s_name = self.clean_s_name(s_name, f)
                parsed = dict()
                for i, v in enumerate(f["f"].split()):
                    parsed[i] = float(v)
                if len(parsed) > 0:
                    if s_name in self.salmon_fld:
                        log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                    self.add_data_source(f, s_name)
                    self.salmon_fld[s_name] = parsed

        # Parse Library Format Counts information. JSON file expected
        self.salmon_lfc = dict()
        for f in self.find_log_files("salmon/lfc"):
            s_name = os.path.basename(f["root"])  # lfc file located at root folder
            s_name = self.clean_s_name(s_name, f)
            self.salmon_lfc[s_name] = json.loads(f["f"])

        # Filter to strip out ignored sample names
        self.salmon_meta = self.ignore_samples(self.salmon_meta)
        self.salmon_fld = self.ignore_samples(self.salmon_fld)
        self.salmon_lfc = self.ignore_samples(self.salmon_lfc)

        if len(self.salmon_meta) == 0 and len(self.salmon_fld) == 0 and len(self.salmon_lfc) == 0:
            raise ModuleNoSamplesFound

        if len(self.salmon_meta) > 0:
            log.info(f"Found {len(self.salmon_meta)} meta reports")
            self.write_data_file(self.salmon_meta, "multiqc_salmon")
        if len(self.salmon_fld) > 0:
            log.info(f"Found {len(self.salmon_fld)} fragment length distributions")
        if len(self.salmon_lfc) > 0:
            log.info(f"Found {len(self.salmon_lfc)} library format counts reports")

        if self.salmon_meta:
            # Add alignment rate to the general stats table
            # Convert library types to string:
            for d in self.salmon_meta.values():
                if "library_types" in d:
                    d["library_types"] = ", ".join(d["library_types"])

            headers = {
                "percent_mapped": {
                    "title": "% Aligned",
                    "description": "% Mapped reads",
                    "max": 100,
                    "min": 0,
                    "suffix": "%",
                    "scale": "YlGn",
                },
                "num_mapped": {
                    "title": "M Aligned",
                    "description": "Mapped reads (millions)",
                    "min": 0,
                    "scale": "PuRd",
                    "modify": lambda x: float(x) * config.read_count_multiplier,
                    "suffix": config.read_count_prefix,
                    "shared_key": "read_count",
                },
                "library_types": {
                    "title": "Library types",
                    "description": "Library types",
                    "scale": False,
                    # Hide if all samples have the same value
                    "hidden": len(set(d.get("library_types") for d in self.salmon_meta.values())) == 1,
                },
            }
            self.general_stats_addcols(self.salmon_meta, headers)

        if self.salmon_lfc:
            # Compatible fragments ratios data
            lfc_headers = {
                "compatible_fragment_ratio": {
                    "title": "CFR",
                    "description": "Compatible fragment ratio",
                    "min": 0.0,
                    "max": 100.0,
                    "modify": lambda x: x * 100.0,
                    "suffix": "%",
                    "scale": "YlGn",
                },
                "strand_mapping_bias": {
                    "title": "M Bias",
                    "description": "Strand mapping bias",
                    "scale": "BuGn",
                    "max": 1.0,
                },
            }
            # add strand mapping bias data
            self.general_stats_addcols(self.salmon_lfc, lfc_headers)

        if self.salmon_fld:
            # Fragment length distribution plot
            pconfig = {
                "smooth_points": 500,
                "id": "salmon_plot",
                "title": "Salmon: Fragment Length Distribution",
                "ylab": "Fraction",
                "xlab": "Fragment Length (bp)",
                "ymin": 0,
                "xmin": 0,
                "tt_label": "<b>{point.x:,.0f} bp</b>: {point.y:,.0f}",
            }
            self.add_section(plot=linegraph.plot(self.salmon_fld, pconfig))



""" MultiQC submodule to parse output from Bamtools bam_stat.py
http://bamtools.sourceforge.net/#bam-stat-py """

import logging
import re

from multiqc.plots import beeswarm
from multiqc.utils import config

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(self):
    """Find bamtools stats reports and parse their data"""

    # Set up vars
    self.bamtools_stats_data = dict()
    regexes = {
        "total_reads": r"Total reads:\s*(\d+)",
        "mapped_reads": r"Mapped reads:\s*(\d+)",
        "mapped_reads_pct": r"Mapped reads:\s*\d+\s+\(([\d\.]+)%\)",
        "forward_strand": r"Forward strand:\s*(\d+)",
        "forward_strand_pct": r"Forward strand:\s*\d+\s+\(([\d\.]+)%\)",
        "reverse_strand": r"Reverse strand:\s*(\d+)",
        "reverse_strand_pct": r"Reverse strand:\s*\d+\s+\(([\d\.]+)%\)",
        "failed_qc": r"Failed QC:\s*(\d+)",
        "failed_qc_pct": r"Failed QC:\s*\d+\s+\(([\d\.]+)%\)",
        "duplicates": r"Duplicates:\s*(\d+)",
        "duplicates_pct": r"Duplicates:\s*\d+\s+\(([\d\.]+)%\)",
        "paired_end": r"Paired-end reads:\s*(\d+)",
        "paired_end_pct": r"Paired-end reads:\s*\d+\s+\(([\d\.]+)%\)",
        "proper_pairs": r"'Proper-pairs'\s*(\d+)",
        "proper_pairs_pct": r"'Proper-pairs'\s*\d+\s+\(([\d\.]+)%\)",
        "both_mapped": r"Both pairs mapped:\s*(\d+)",
        "both_mapped_pct": r"Both pairs mapped:\s*\d+\s+\(([\d\.]+)%\)",
        "read_1": r"Read 1:\s*(\d+)",
        "read_2": r"Read 2:\s*(\d+)",
        "singletons": r"Singletons:\s*(\d+)",
        "singletons_pct": r"Singletons:\s*\d+\s+\(([\d\.]+)%\)",
    }

    # Go through files and parse data using regexes
    for f in self.find_log_files("bamtools/stats"):
        d = dict()
        for k, r in regexes.items():
            r_search = re.search(r, f["f"], re.MULTILINE)
            if r_search:
                d[k] = float(r_search.group(1))

        if len(d) > 0:
            if f["s_name"] in self.bamtools_stats_data:
                log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
            self.add_data_source(f, section="stats")
            self.bamtools_stats_data[f["s_name"]] = d

    # Filter to strip out ignored sample names
    self.bamtools_stats_data = self.ignore_samples(self.bamtools_stats_data)

    if len(self.bamtools_stats_data) == 0:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    self.add_software_version(None)

    # Write to file
    self.write_data_file(self.bamtools_stats_data, "multiqc_bamtools_stats")

    # Add to general stats table
    self.general_stats_headers["duplicates_pct"] = {
        "title": "% Duplicates",
        "description": "% Duplicate Reads",
        "max": 100,
        "min": 0,
        "suffix": "%",
        "scale": "OrRd",
    }
    self.general_stats_headers["mapped_reads_pct"] = {
        "title": "% Mapped",
        "description": "% Mapped Reads",
        "max": 100,
        "min": 0,
        "suffix": "%",
        "scale": "RdYlGn",
    }
    for s_name in self.bamtools_stats_data:
        if s_name not in self.general_stats_data:
            self.general_stats_data[s_name] = dict()
        self.general_stats_data[s_name].update(self.bamtools_stats_data[s_name])

    # Make dot plot of counts
    keys = {}
    defaults = {
        "min": 0,
        "max": 100,
        "decimalPlaces": 2,
        "suffix": "%",
    }
    num_defaults = {
        "min": 0,
        "modify": lambda x: x * config.read_count_multiplier,
        "suffix": config.read_count_prefix,
        "decimalPlaces": 2,
    }

    keys["total_reads"] = dict(num_defaults, **{"title": "Total reads", "description": "Total reads (millions)"})
    keys["mapped_reads_pct"] = dict(defaults, **{"title": "Mapped reads"})
    keys["forward_strand_pct"] = dict(defaults, **{"title": "Forward strand"})
    keys["reverse_strand_pct"] = dict(defaults, **{"title": "Reverse strand"})
    keys["failed_qc_pct"] = dict(defaults, **{"title": "Failed QC"})
    keys["duplicates_pct"] = dict(defaults, **{"title": "Duplicates"})
    keys["paired_end_pct"] = dict(defaults, **{"title": "Paired-end", "description": "Paired-end reads"})
    keys["proper_pairs_pct"] = dict(defaults, **{"title": "Proper-pairs"})
    keys["both_mapped_pct"] = dict(defaults, **{"title": "Both mapped", "description": "Both pairs mapped"})
    keys["bt_read_1"] = dict(num_defaults, **{"title": "Read 1", "description": "Read 1 (millions)"})
    keys["bt_read_2"] = dict(num_defaults, **{"title": "Read 2", "description": "Read 2 (millions)"})
    keys["singletons_pct"] = dict(defaults, **{"title": "Singletons"})

    self.add_section(
        name="Bamtools Stats",
        anchor="bamtools-stats",
        plot=beeswarm.plot(
            self.bamtools_stats_data,
            keys,
            pconfig={
                "id": "bamtools-stats-plot",
                "title": "Bamtools Stats",
            },
        ),
    )

    # Return number of samples found
    return len(self.bamtools_stats_data)



import logging
import re

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import table

log = logging.getLogger(__name__)

VERSION_REGEX = r"# Version: v([\d\.]+)"


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Illumina InterOp Statistics",
            anchor="interop",
            href="http://illumina.github.io/interop/index.html",
            info=" - a set of common routines used for reading and writing InterOp metric files.",
            # No publication / DOI // doi=
        )

        log = logging.getLogger(__name__)

        # Parse data
        self.runSummary = {}
        self.indexSummary = {}
        for f in self.find_log_files("interop/summary", filehandles=True):
            parsed_data, version = self.parse_summary_csv(f["f"])
            if max(len(parsed_data["summary"]), len(parsed_data["details"])) > 0:
                self.runSummary[f["s_name"]] = parsed_data
                self.add_data_source(f)

            if version is not None:
                self.add_software_version(version, f["s_name"])

        for f in self.find_log_files("interop/index-summary", filehandles=True):
            parsed_data, version = self.parse_index_summary_csv(f["f"])
            if max(len(parsed_data["summary"]), len(parsed_data["details"])) > 0:
                self.indexSummary[f["s_name"]] = parsed_data
                self.add_data_source(f)

            if version is not None:
                self.add_software_version(version, f["s_name"])

        self.runSummary = self.ignore_samples(self.runSummary)
        self.indexSummary = self.ignore_samples(self.indexSummary)

        # No samples
        num_samples = max(len(self.runSummary), len(self.indexSummary))
        if num_samples == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {num_samples} reports")

        # Create report Sections
        if len(self.runSummary) > 0:
            # Write data to file
            self.write_data_file(self.runSummary, "interop_runsummary")

            self.add_section(
                name="Read Metrics Summary",
                anchor="interop-runmetrics-summary",
                description="Summary statistics for Total read count from each run.",
                plot=self.run_metrics_summary_table(self.runSummary),
            )
            self.add_section(
                name="Read Metrics per Lane",
                anchor="interop-runmetrics-details",
                plot=self.run_metrics_details_table(self.runSummary),
            )

        if len(self.indexSummary) > 0:
            # Write data to file
            self.write_data_file(self.indexSummary, "interop_indexsummary")

            self.add_section(
                name="Indexing QC Metrics summary",
                anchor="interop-indexmetrics-summary",
                description="Summary metrics about each lane",
                plot=self.index_metrics_summary_table(self.indexSummary),
            )
            self.add_section(
                name="Indexing QC Metrics details",
                anchor="interop-indexmetrics-details",
                description=" Detail Metrics about each lane",
                plot=self.index_metrics_details_table(self.indexSummary),
            )

    @staticmethod
    def parse_summary_csv(f):
        metrics = {"summary": {}, "details": {}}
        header = []
        section = None
        read = None
        version = None
        for line in f:
            line = line.strip()
            # assume fixed file format

            # find version
            if line.startswith("# Version"):
                match = re.search(VERSION_REGEX, line)
                if match:
                    version = match.group(1)

            # find summary header
            if line.startswith("Level,Yield,Projected Yield,Aligned,Error Rate,Intensity C1,%>=Q30"):
                # set section to summary
                section = "summary"
                header = line.split(",")
            elif section == "summary":
                data = line.split(",")
                # process summary
                metrics["summary"][data[0]] = {}
                for idx in range(1, len(data)):
                    try:
                        metrics["summary"][data[0]][header[idx]] = int(data[idx])
                    except ValueError:
                        try:
                            metrics["summary"][data[0]][header[idx]] = float(data[idx])
                        except ValueError:
                            metrics["summary"][data[0]][header[idx]] = data[idx]
                if line.startswith("Total"):
                    section = None
            elif line.startswith("Read") and (section is None or section == "details"):
                # set section to details
                section = "details"
                read = line
            elif line.startswith("Lane,Surface,Tiles,Density,Cluster") and section == "details":
                # get details header
                header = line.split(",")
            elif section == "details":
                if line.startswith("Extracted: "):
                    section = "finish"
                    continue
                data = line.split(",")
                # process summary
                linedata = {}
                # Check if "surface" is total (-) else skip
                if data[1] == "-":
                    metrics["details"][f"Lane {data[0]} - {read}"] = {}
                else:
                    continue
                for idx in range(2, len(data)):
                    try:
                        if header[idx] == "Phas/Prephas":
                            val = data[idx].split("/")
                            linedata["Phased"] = val[0]
                            linedata["Prephased"] = val[1]
                        elif header[idx] == "Cycles Error":
                            vals = data[idx].split(" - ")
                            linedata[header[idx]] = max(int(v) for v in vals)
                        else:
                            try:
                                linedata[header[idx]] = int(data[idx])
                            except ValueError:
                                linedata[header[idx]] = float(data[idx])
                    except ValueError:
                        val = re.sub(pattern=r"\+/-.*", repl="", string=data[idx]).strip()
                        try:
                            linedata[header[idx]] = int(val)
                        except ValueError:
                            try:
                                linedata[header[idx]] = float(val)
                            except ValueError:
                                linedata[header[idx]] = val
                metrics["details"][f"Lane {data[0]} - {read}"] = linedata

        return metrics, version

    @staticmethod
    def parse_index_summary_csv(f):
        metrics = {"summary": {}, "details": {}}
        header = []
        section = None
        lane = None

        summary = {}
        details = {}
        version = None
        for line in f:
            line = line.strip()
            # assume fixed file format
            # find version
            if line.startswith("# Version"):
                match = re.search(VERSION_REGEX, line)
                if match:
                    version = match.group(1)

            if line.startswith("Lane"):
                # set lane
                lane = line
                summary[lane] = {}
                continue
            if line.startswith("Total Reads,PF Reads,% Read Identified (PF),CV,Min,Max"):
                header = line.split(",")
                section = "summary"
                continue
            if line.startswith("Index Number,Sample Id,Project,Index 1 (I7),Index 2 (I5),% Read Identified (PF)"):
                header = line.split(",")
                section = "details"
                continue
            if section == "summary":
                data = line.split(",")
                for idx in range(0, len(data)):
                    summary[lane][header[idx]] = data[idx]
                continue
            if section == "details":
                data = line.split(",")
                details[f"{data[1]} - {lane}"] = {}
                for idx in range(2, len(data)):
                    details[f"{data[1]} - {lane}"][header[idx]] = data[idx]
                continue

        metrics["summary"] = summary
        metrics["details"] = details

        return metrics, version

    @staticmethod
    def run_metrics_summary_table(data):
        headers = {
            "Yield": {
                "rid": "summary_Yield",
                "title": "Gbp Yield",  # Numbers are rounded up to Gbp, so no point in using multiplier for smaller numbers as will be all zeroes
                "description": 'The number of bases sequenced (Gbp base pairs over all "usable cycles")',
                "scale": "PuOr",
                "min": 0,
                "format": "{:,.2f}",
            },
            "Aligned": {
                "rid": "summary_Aligned",
                "title": "Aligned (%)",
                "description": "The percentage of the sample that aligned to the PhiX genome",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "PiYG",
            },
            "Error Rate": {
                "title": "Error Rate (%)",
                "description": "",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "OrRd",
            },
            "Intensity C1": {
                "rid": "summary_Intensity_C1",
                "title": "Intensity Cycle 1",
                "description": "The intensity statistic at cycle 1.",
                "min": 0,
                "scale": "PuOr",
                "format": "{:,d}",
            },
            "%>=Q30": {
                "rid": "summary_Q30",
                "title": "% >= Q30",
                "description": "Percentage of reads with quality phred score of 30 or above",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "RdYlGn",
            },
        }
        table_config = {
            "namespace": "interop",
            "id": "interop-runmetrics-summary-table",
            "table_title": "Read metrics summary",
            "col1_header": "Run - Read",
        }

        tdata = {}
        for s_name in data:
            for key in data[s_name]["summary"]:
                tdata[f"{s_name} - {key}"] = data[s_name]["summary"][key]

        return table.plot(tdata, headers, table_config)

    @staticmethod
    def run_metrics_details_table(data):
        headers = {
            "Surface": {"title": "Surface", "description": ""},
            "Tiles": {"title": "Tiles", "description": "The number of tiles per lane.", "hidden": True},
            "Density": {
                "title": "Density",
                "description": "The density of clusters (in thousands per mm2) detected by image analysis, +/- 1 standard deviation.",
                "hidden": True,
            },
            "Cluster PF": {
                "title": "Cluster PF (%)",
                "description": "The percentage of clusters passing filtering, +/- 1 standard deviation.",
                "suffix": "%",
            },
            "% Occupied": {
                "title": "Occupied (%)",
                "description": "The percentage of nanowells occupied by clusters, +/- 1 standard deviation.",
                "suffix": "%",
            },
            "Phased": {
                "title": "Phased (%)",
                "description": "The value used by RTA for the percentage of molecules in a cluster for which sequencing falls behind (phasing) or jumps ahead (prephasing) the current cycle within a read.",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "OrRd",
            },
            "Prephased": {
                "title": "Prephased (%)",
                "description": "The value used by RTA for the percentage of molecules in a cluster for which sequencing falls behind (phasing) or jumps ahead (prephasing) the current cycle within a read.",
                "format": "{:.,2f}",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "OrRd",
            },
            "Reads": {
                "title": "M Reads",
                "description": "The number of clusters (millions)",
            },
            "Reads PF": {
                "title": "M PF Reads",
                "description": "The number of passing filter clusters (millions)",
            },
            "Cycles Error": {
                "title": "Cycles Error",
                "description": "The number of cycles that have been error-rated using PhiX, starting at cycle 1.",
                "format": "{:.,0f}",
                "scale": "OrRd",
            },
            "Yield": {
                "title": "Gbp Yield",
                "description": "The number of bases sequenced which passed filter (Gbp base pairs)",
                "scale": "PuOr",
                "min": 0,
                "format": "{:,.2f}",
            },
            "Aligned": {
                "title": "Aligned (%)",
                "description": "The percentage that aligned to the PhiX genome.",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "PiYG",
            },
            "Error": {
                "title": "Error Rate (%)",
                "description": "The calculated error rate, as determined by the PhiX alignment.",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "OrRd",
            },
            "Error (35)": {
                "title": "Error Rate 35 Cycles (%)",
                "description": "The calculated error rate for cycles 1-35.",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "OrRd",
                "hidden": True,
            },
            "Error (75)": {
                "title": "Error Rate 75 Cycles (%)",
                "description": "The calculated error rate for cycles 1-75.",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "OrRd",
                "hidden": True,
            },
            "Error (100)": {
                "title": "Error Rate 100 Cycles (%)",
                "description": "The calculated error rate for cycles 1-100.",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "OrRd",
                "hidden": True,
            },
            "Intensity C1": {
                "title": "Intensity Cycle 1",
                "description": "The intensity statistic at cycle 1.",
                "min": 0,
                "scale": "PuOr",
                "format": "{:,d}",
            },
            "%>=Q30": {
                "title": "%>=Q30",
                "description": "The percentage of bases with a quality score of 30 or higher, respectively.",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "RdYlGn",
            },
        }
        table_config = {
            "namespace": "interop",
            "id": "interop-runmetrics-detail-table",
            "table_title": "Sequencing Lane Statistics",
            "col1_header": "Run - Lane - Read",
        }

        tdata = {}
        for s_name in data:
            for key in data[s_name]["details"]:
                tdata[f"{s_name} - {key}"] = data[s_name]["details"][key]

        return table.plot(tdata, headers, table_config)

    @staticmethod
    def index_metrics_summary_table(data):
        headers = {
            "Total Reads": {
                "rid": "interop_reads_total",
                "title": f"{config.read_count_prefix} Reads",
                "description": f"The total number of reads for this lane ({config.read_count_desc})",
                "modify": lambda x: float(x) * config.read_count_multiplier,
                "format": "{:,.2f}",
                "shared_key": "read_count",
            },
            "PF Reads": {
                "title": f"{config.read_count_prefix} PF Reads",
                "description": "The total number of passing filter reads for this lane ({})".format(
                    config.read_count_desc
                ),
                "modify": lambda x: float(x) * config.read_count_multiplier,
                "format": "{:,.2f}",
                "shared_key": "read_count",
            },
            "% Read Identified (PF)": {
                "rid": "summary_reads_identified_pf",
                "title": "% Reads Identified (PF)",
                "description": "The total fraction of passing filter reads assigned to an index.",
                "suffix": "%",
            },
            "CV": {
                "title": "CV",
                "description": "The coefficient of variation for the number of counts across all indexes.",
                "format": "{:.,2f}",
            },
            "Min": {"title": "Min", "description": "The lowest representation for any index."},
            "Max": {"title": "Max", "description": "The highest representation for any index."},
        }
        table_config = {
            "namespace": "interop",
            "id": "interop-indexmetrics-summary-table",
            "table_title": "Index Read Statistics Summary",
            "col1_header": "Run - Lane",
        }

        tdata = {}
        for s_name in data:
            for key in data[s_name]["summary"]:
                tdata[f"{s_name} - {key}"] = data[s_name]["summary"][key]

        return table.plot(tdata, headers, table_config)

    @staticmethod
    def index_metrics_details_table(data):
        headers = {
            "% Read Identified (PF)": {
                "title": "% Reads Identified (PF)",
                "description": "The number of reads (only includes Passing Filter reads) mapped to this index.",
                "suffix": "%",
            },
            "Index 1 (I7)": {"title": "Index 1 (I7)", "description": "The sequence for the first Index Read."},
            "Index 2 (I5)": {"title": "Index 2 (I5)", "description": "The sequence for the second Index Read."},
        }

        table_config = {
            "namespace": "interop",
            "id": "interop-indexmetrics-details-table",
            "table_title": "Index Read Statistics Details",
            "col1_header": "Run - Sample - Lane",
        }

        tdata = {}
        for s_name in data:
            for key in data[s_name]["details"]:
                tdata[f"{s_name} - {key}"] = data[s_name]["details"][key]

        return table.plot(tdata, headers, table_config)



from itertools import chain

from multiqc.plots import linegraph


def plot_idhist(samples, file_type, **plot_args):
    """Create line graph plot of histogram data for BBMap 'idhist' output.

    The 'samples' parameter could be from the bbmap mod_data dictionary:
    samples = bbmap.MultiqcModule.mod_data[file_type]
    """

    all_x = set()
    for item in sorted(chain(*[samples[sample]["data"].items() for sample in samples])):
        all_x.add(item[0])

    columns_to_plot = {
        "Reads": {
            0: "Count",
        },
        "Bases": {
            1: "Count",
        },
    }

    plot_data = []
    for column_type in columns_to_plot:
        plot_data.append(
            {
                sample + "." + column_name: {
                    x: samples[sample]["data"][x][column] if x in samples[sample]["data"] else 0 for x in all_x
                }
                for sample in samples
                for column, column_name in columns_to_plot[column_type].items()
            }
        )

    plot_params = {
        "id": "bbmap-" + file_type + "_plot",
        "title": "BBTools: " + plot_args["plot_title"],
        "xlab": "Percent identity",
        "ylab": "Read count",
        "data_labels": [
            {"name": "Reads", "ylab": "Read count"},
            {"name": "Bases", "ylab": "Number of bases"},
        ],
    }
    plot_params.update(plot_args["plot_params"])
    plot = linegraph.plot(plot_data, plot_params)

    return plot



""" MultiQC module to parse output from Supernova"""


import json
import logging
import re

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, linegraph, table

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        super(MultiqcModule, self).__init__(
            name="Supernova",
            anchor="supernova",
            href="https://www.10xgenomics.com/",
            info="is a de novo genome assembler 10X Genomics linked-reads.",
            doi="10.1101/gr.214874.116",
        )

        # Headers for the supernova Table
        headers = {
            "Asm size": {
                "description": "assembly size (in megabases) ;only scaffolds >= 10 kb",
                "modify": lambda x: x / 1000000.0,
                "suffix": "Mb",
                "scale": "YlGn",
            },
            "% missing 10Kb": {
                "rid": "pct_missing_10Kb",
                "description": "% of base assembly missing from scaffolds >= 10 kb",
                "suffix": "%",
                "scale": "YlGn",
            },
            "# Long scaffs": {
                "rid": "num_long_scaffs",
                "description": "number of scaffolds >= 10 kb",
                "scale": "YlGn",
                "format": "{:,.0f}",
                "hidden": True,
            },
            "Scaff N50": {
                "description": "N50 scaffold size (in kilobases)",
                "modify": lambda x: x / 1000.0,
                "suffix": "Kb",
                "scale": "RdYlGn",
            },
            "Phase N50": {
                "description": "N50 phase block size (in kilobases)",
                "modify": lambda x: x / 1000.0,
                "suffix": "Kb",
                "scale": "RdYlGn",
                "hidden": True,
            },
            "Contig N50": {
                "description": "N50 contig size (in kilobases)",
                "modify": lambda x: x / 1000.0,
                "suffix": "Kb",
                "scale": "RdYlGn",
                "hidden": True,
            },
            "Edge N50": {
                "description": "N50 edge size (in kilobases)",
                "modify": lambda x: x / 1000.0,
                "suffix": "Kb",
                "scale": "RdYlGn",
                "hidden": True,
            },
            "Mol size": {
                "description": "weighted mean molecule size (in kilobases); ideal 50-100",
                "modify": lambda x: x / 1000.0,
                "suffix": "Kb",
                "scale": "BuGn",
            },
            "Read len": {
                "description": "mean read length (in bases) after trimming; ideal 140",
                "suffix": "b",
                "scale": "PuBu",
                "format": "{:,.0f}",
                "hidden": True,
            },
            "# Reads": {
                "rid": "num_reads",
                "description": "number of reads (in millions); ideal 800M-1200M for human",
                "modify": lambda x: x / 1000000.0,
                "suffix": "M",
                "scale": "PuBu",
            },
            "Raw coverage": {
                "description": "raw coverage; ideal ~56",
                "suffix": "x",
                "scale": "PuBu",
                "hidden": True,
            },
            "Coverage": {
                "description": "effective read coverage; ideal ~42 for nominal 56x cov",
                "suffix": "x",
                "scale": "PuBu",
            },
            "% Dup": {
                "rid": "pct_Dup",
                "description": "fraction of reads that are duplicates",
                "suffix": "%",
                "scale": "OrRd",
            },
            "% R2 Q30": {
                "rid": "pct_R2_Q30",
                "description": "fraction of Q30 bases in read 2; ideal 75-85%",
                "suffix": "%",
                "scale": "OrRd",
            },
            "Insert size": {
                "description": "median insert size (in bases); ideal 0.35-0.40 Kb",
                "suffix": "b",
                "scale": "OrRd",
                "format": "{:,.0f}",
                "hidden": True,
            },
            "% proper": {
                "rid": "pct_proper",
                "description": "fraction of proper read pairs; ideal >= 75%",
                "suffix": "%",
                "scale": "OrRd",
                "hidden": True,
            },
            "BC usage": {
                "description": "fraction of barcodes used; between 0 and 1",
                "scale": "OrRd",
                "hidden": True,
            },
            "Est size": {
                "description": "estimated genome size",
                "modify": lambda x: x / 1000000.0,
                "suffix": "Mb",
                "scale": "YlGn",
                "hidden": True,
            },
            "% repeats": {
                "rid": "pct_repeats",
                "description": "Estimated repetitive fraction (of genome)",
                "scale": "YlGn",
                "suffix": "%",
                "hidden": True,
            },
            "% AT": {
                "rid": "pct_AT",
                "description": "high AT index (of genome)",
                "scale": "YlGn",
                "suffix": "%",
                "hidden": True,
            },
            "Het dist": {
                "description": "mean distance between heterozygous SNPs (in kilobases)",
                "modify": lambda x: x / 1000.0,
                "suffix": "Kb",
                "scale": "YlGn",
                "format": "{:,.0f}",
                "hidden": True,
            },
            "p10": {
                "description": "molecule count extending 10 kb on both sides",
                "scale": "BuGn",
                "hidden": True,
            },
            "% missing BC": {
                "rid": "pct_missing_BC",
                "description": "fraction of reads that are not barcoded",
                "suffix": "%",
                "scale": "BuGn",
            },
            "Barcode N50": {
                "description": "N50 reads per barcode (in bases)",
                "suffix": "b",
                "scale": "BuGn",
                "format": "{:,.0f}",
            },
            "% Phased": {
                "rid": "pct_Phased",
                "description": "nonduplicate and phased reads; ideal 45-50%",
                "suffix": "%",
                "scale": "BuGn",
                "hidden": True,
            },
        }

        reports = {}
        summaries = {}
        molecules = {}
        kmers = {}
        root_summary = {}

        # Parse the input log files
        # report.txt files
        for f in self.find_log_files("supernova/report"):
            log.debug(f"Found report in: {f['root']}")
            sid, data = self.parse_report(f["f"])
            s_name = self.clean_s_name(sid, f)
            if s_name in reports.keys():
                log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
            reports[s_name] = data
            self.add_data_source(f, s_name=s_name, section="supernova-table")

        # summary.json files
        for f in self.find_log_files("supernova/summary"):
            log.debug(f"Found summary.json in: {f['root']}")
            try:
                sid, data = self.parse_summary(f["f"])
            except ValueError:
                log.debug(f"Error parsing JSON file in {f['root']}")
                continue
            except RuntimeError:
                log.debug(f"Could not find sample_id in JSON file in {f['root']}")
                continue

            s_name = self.clean_s_name(sid, f)
            if s_name in summaries.keys():
                log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
            summaries[s_name] = data
            self.add_data_source(f, s_name=s_name, section="supernova-table")
            # The plot json files do not contain sample IDs, sadly. So we need to store it somewhere.
            root_summary[f["root"]] = sid

        # histogram_molecules.json files
        for f in self.find_log_files("supernova/molecules"):
            log.debug(f"Found histogram_molecules.json in: {f['root']}")
            try:
                if f["root"] in root_summary.keys():
                    data = self.parse_histogram(f["f"])
                    sid = root_summary[f["root"]]
                    s_name = self.clean_s_name(sid, f)
                    molecules[s_name] = data
                    self.add_data_source(f, s_name=s_name, section="supernova-molecules")
            except RuntimeError:
                log.debug(f"Could not parse JSON file in {f['root']}")
                continue

        # histogram_kmer_count.json files
        for f in self.find_log_files("supernova/kmers"):
            log.debug(f"Found histogram_kmer_count.json in: {f['root']}")
            try:
                if f["root"] in root_summary.keys():
                    data = self.parse_histogram(f["f"], 400)
                    sid = root_summary[f["root"]]
                    s_name = self.clean_s_name(sid, f)
                    kmers[s_name] = data
                    self.add_data_source(f, s_name=s_name, section="supernova-kmers")
            except RuntimeError:
                log.debug(f"Could not parse JSON file in {f['root']}")
                continue

        # Data from summary.json supersedes data from report.txt
        for sample_id, sum_data in summaries.items():
            if sample_id in reports.keys():
                log.debug(f"Found summary data for sample {sample_id} which supersedes report data")
                reports[sample_id] = sum_data
        # Ignore cmd-line specified samples
        reports = self.ignore_samples(reports)
        molecules = self.ignore_samples(molecules)
        kmers = self.ignore_samples(kmers)

        if len(reports) == 0:
            raise ModuleNoSamplesFound
        else:
            log.info(f"Found {len(reports.keys())} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write the report
        self.write_data_file(reports, "multiqc_supernova")
        config_table = {"id": "supernova_table", "namespace": "supernova", "title": "Supernova: Assembly statistics"}
        self.add_section(
            name="Assembly statistics",
            anchor="supernova-table",
            description="Statistics gathered from the summary report(s) of Supernova. Note! "
            "There are more columns available but they are hidden by default.",
            helptext="As a bare minimum these numbers are generated from the file report.txt, "
            "found in the folder `sampleID/outs/`. If available the stats in the report "
            "file will be superseded by the higher precision numbers found in the file "
            "`sampleID/outs/assembly/stats/summary.json`",
            plot=table.plot(reports, headers, config_table),
        )

        # N50 barcharts
        n50_cats = [
            {"Scaff N50": {"name": "Scaffold N50", "color": "#66c2a5"}},
            {"Contig N50": {"name": "Contig N50", "color": "#fc8d62"}},
            {"Edge N50": {"name": "Edge N50", "color": "#8da0cb"}},
            {"Phase N50": {"name": "Phase block N50", "color": "#e78ac3"}},
        ]
        config_n50 = {
            "id": "supernova_n50",
            "title": "Supernova: N50 statistics",
            "ylab": "Scaffold N50",
            "cpswitch": False,
            "data_labels": ["Scaffold N50", "Contig N50", "Edge N50", "Phase block N50"],
        }
        self.add_section(
            name="N50 statistics",
            anchor="supernova-n50",
            description="Assembly N50 values - the shortest sequence length at 50% of the genome when sorted by size (see [wikipedia](https://en.wikipedia.org/wiki/N50,_L50,_and_related_statistics#N50)).",
            helptext="Note that assembly size and N50 values are computed after removing scaffolds &le; 10 kb and do not count `N`s: \n\n"
            "* **Scaffold N50** - N50 size of scaffolds in bases, \n"
            "* **Contig N50** - N50 size of contigs in bases, \n"
            "* **Edge N50** - N50 size of raw graph assembly edges in bases, \n"
            "* **Phase block N50** - N50 size of phase blocks in bases. \n\n"
            "[(source)](https://support.10xgenomics.com/de-novo-assembly/software/pipelines/latest/output/asm-stats)",
            plot=bargraph.plot([reports, reports, reports, reports], n50_cats, config_n50),
        )

        # Conditional sections
        if len(molecules) > 0:
            # Remove the long tail, or fail if this is a legacy empty json file
            try:
                max_x = self.trim_tail(molecules, 100000)
            except IndexError:
                log.debug("The histogram file is empty. Skipping molecule length section")
                return
            # Add molecules plot
            config_molecules = {
                "id": "supernova_molecules",
                "title": "Supernova: Molecule Lengths",
                "xlab": "Inferred molecule length (bp)",
                "ylab": "# molecules",
                "smooth_points": 300,
                "smooth_points_sumcounts": True,
                "xmax": max_x,
            }
            self.add_section(
                name="Molecule Lengths",
                anchor="supernova-molecules",
                description="Shows the inferred molecule lengths of the input 10X library.",
                helptext="Inferred in the `patch` step of the Supernova pipeline. It is worth "
                "keeping in mind that the mean molecule length from the report is a length-weighted mean. "
                "See the [source code](https://github.com/10XGenomics/supernova/search?q=lw_mean_mol_len&type=) "
                "for how this value is calculated.",
                plot=linegraph.plot(molecules, config_molecules),
            )
        if len(kmers) > 0:
            # Remove the long tail, or fail if this is a legacy empty json file
            try:
                max_x = self.trim_tail(kmers, 50)
            except IndexError:
                log.debug("The histogram file is empty. Skipping kmers section")
                return
            # Add kmers plot
            config_kmers = {
                "id": "supernova_kmers",
                "title": "Supernova: Kmer Counts",
                "xlab": "Filtered kmer multiplicity",
                "ylab": "Counts",
                "smooth_points_sumcounts": False,
                "xmax": max_x,
            }
            self.add_section(
                name="K-mer counts",
                anchor="supernova-kmers",
                description="Shows the k-mer frequencies of the input data to Supernova (after filtering).",
                helptext="This data is generated from k-merizing the input read data, where the sequences are "
                "transformed in to the set of all possible sub-sequences of a fixed length of `K` (Supernova uses `K=48`). "
                "The plot shows on the x-axis the multiplicity (i.e. how many times are they repeated) of these k-mers "
                "and the y-axis the number of k-mers at this level of multiplicity. "
                "A careful reading of this plot can give some insights into the levels of heterozygosity and repeats "
                "in the genome that was sequenced and indications if the sequencing experiment was successful.",
                plot=linegraph.plot(kmers, config_kmers),
            )

    @staticmethod
    def parse_summary(content):
        stats = {
            "assembly_size": "Asm size",
            "bases_per_read": "Read len",
            "contig_N50": "Contig N50",
            "dup_perc": "% Dup",
            "edge_N50": "Edge N50",
            "effective_coverage": "Coverage",
            "hetdist": "Het dist",
            "lw_mean_mol_len": "Mol size",
            "median_ins_sz": "Insert size",
            "nreads": "# Reads",
            "phase_block_N50": "Phase N50",
            "placed_perc": "% Phased",
            "placed_frac": "% Phased",
            "proper_pairs_perc": "% proper",
            "q30_r2_perc": "% R2 Q30",
            "rpb_N50": "Barcode N50",
            "scaffold_N50": "Scaff N50",
            "scaffolds_10kb_plus": "# Long scaffs",
            "valid_bc_perc": "% missing BC",
            "m10": "% missing 10Kb",
            "high_AT_index": "% AT",
            "raw_coverage": "Raw coverage",
            "barcode_fraction": "BC usage",
            "repfrac": "Repeats",
            "est_genome_size": "Est size",
            "p10": "p10",
        }

        try:
            cdict = json.loads(content)
        except ValueError as e:
            raise e

        data = {}
        # Try to find sample_id
        sid = ""
        if "CS_SAMPLE_ID" in cdict.keys():
            sid = cdict["CS_SAMPLE_ID"]  # supernova 1.2
        elif "sample_id" in cdict.keys():
            sid = cdict["sample_id"]
        else:
            raise RuntimeError

        for key, value in cdict.items():
            if key in stats.keys():
                # Some trickery for supernova 1.1.4 compatability
                if key == "placed_frac":
                    value = value * 100
                if key == "valid_bc_perc":
                    value = 100 - value
                data[stats[key]] = value

        return sid, data

    @staticmethod
    def parse_report(content):
        # Some short-hands for converting the report numbers (pi is exactly three!)
        exp = {
            "K": 1000.0,
            "Kb": 1000.0,
            "kb": 1000.0,
            "Mb": 1000000.0,
            "M": 1000000.0,
            "Gb": 1000000000.0,
            "G": 1000000000.0,
        }
        stats = {
            "READS": "# Reads",
            "MEAN READ LEN": "Read len",
            "EFFECTIVE COV": "Coverage",
            "READ TWO Q30": "% R2 Q30",
            "MEDIAN INSERT": "Insert size",
            "PROPER PAIRS": "% proper",
            "MOLECULE LEN": "Mol size",
            "HETDIST": "Het dist",
            "UNBAR": "% missing BC",
            "BARCODE N50": "Barcode N50",
            "DUPS": "% Dup",
            "PHASED": "% Phased",
            "LONG SCAFFOLDS": "# Long scaffs",
            "EDGE N50": "Edge N50",
            "CONTIG N50": "Contig N50",
            "PHASEBLOCK N50": "Phase N50",
            "SCAFFOLD N50": "Scaff N50",
            "ASSEMBLY SIZE": "Asm size",
            "MISSING 10KB": "% missing 10Kb",
            "HIGH AT FRACTION": "% AT",
            "RAW COV": "Raw coverage",
            "BARCODE FRACTION": "BC usage",
            "REPETITIVE FRAC": "Repeats",
            "EST GENOME SIZE": "Est size",
            "P10": "p10",
        }

        data = {}
        # Find the sample ID
        sid = ""
        sid_pat = re.compile(r"- \[(.+)\]")
        # [number, unit, category]
        stat_pat = re.compile(r"-\s+(\d+\.\d+)\s+(\S+|.)\s+= (.+) =")

        for line in content.splitlines():
            sid_m = re.match(sid_pat, line)
            stat_m = re.match(stat_pat, line)

            if sid_m is not None:
                sid = sid_m.groups()[0]
            if stat_m is None:
                continue
            stat_val = stat_m.groups()
            stat_type = stat_val[2].strip()
            # Parse the lines containing statistics
            if stat_type in stats.keys():
                try:
                    if stat_val[1] in exp.keys():
                        data[stats[stat_type]] = float(stat_val[0]) * exp[stat_val[1]]
                    else:
                        data[stats[stat_type]] = float(stat_val[0])
                except ValueError:
                    log.debug(f'Error in parsing sample {sid}, on line "{stat_val}"')

        return sid, data

    @staticmethod
    def parse_histogram(content, cutoff=None):
        try:
            cdict = json.loads(content)
        except ValueError as e:
            raise e

        numbins = cdict["numbins"] + 1
        xdata = [i * cdict["binsize"] for i in range(0, numbins)]
        return {i: j for (i, j) in zip(xdata, cdict["vals"][:cutoff])}

    @staticmethod
    def trim_tail(plot, min_x=50, pct=0.99):
        join_plot = {}
        cuml_plot = {}
        for sample, plot_data in plot.items():
            for key, value in plot_data.items():
                join_plot[key] = join_plot.get(key, 0) + value
        max_i = 0
        for key in join_plot.keys():
            max_i += join_plot[key]
            cuml_plot[key] = max_i
        max_x = [i for i, j in cuml_plot.items() if j <= max_i * pct][-1]
        # xlim = {, 50} at minimum
        if max_x < min_x:
            max_x = min_x
        return max_x



""" MultiQC module to parse logs from SnpEff """


import logging
import re

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, linegraph

# Initialise the logger
log = logging.getLogger(__name__)

VERSION_REGEX = r"SnpEff_version , SnpEff ([\d\.a-z]+)"


class MultiqcModule(BaseMultiqcModule):
    """SnpEff"""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="SnpEff",
            anchor="snpeff",
            href="http://snpeff.sourceforge.net/",
            info=" is a genetic variant annotation and effect prediction "
            "toolbox. It annotates and predicts the effects of variants "
            "on genes (such as amino acid changes). ",
            doi="10.4161/fly.19695",
        )

        self.snpeff_data = dict()
        self.snpeff_section_totals = dict()
        self.snpeff_qualities = dict()

        for f in self.find_log_files("snpeff", filehandles=True):
            self.parse_snpeff_log(f)

        # Filter to strip out ignored sample names
        self.snpeff_data = self.ignore_samples(self.snpeff_data)

        if len(self.snpeff_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.snpeff_data)} reports")

        # Write parsed report data to a file
        self.write_data_file(self.snpeff_data, "multiqc_snpeff")

        # General stats table
        self.general_stats()

        # Report sections
        self.add_section(
            name="Variants by Genomic Region",
            description=(
                """
                The stacked bar plot shows locations of detected variants in
                the genome and the number of variants for each location.
                """
            ),
            helptext=(
                """
                The upstream and downstream interval size to detect these
                genomic regions is 5000bp by default.
                """
            ),
            anchor="snpeff-genomic-regions",
            plot=self.count_genomic_region_plot(),
        )
        self.add_section(
            name="Variant Effects by Impact",
            description=(
                """
                The stacked bar plot shows the putative impact of detected
                variants and the number of variants for each impact.
                """
            ),
            helptext=(
                """
                There are four levels of impacts predicted by SnpEff:

                * **High**: High impact (like stop codon)
                * **Moderate**: Middle impact (like same type of amino acid substitution)
                * **Low**: Low impact (ie silence mutation)
                * **Modifier**: No impact
                """
            ),
            anchor="snpeff-effects-impact",
            plot=self.effects_impact_plot(),
        )
        self.add_section(
            name="Variants by Effect Types",
            description=(
                """
                The stacked bar plot shows the effect of variants at protein
                level and the number of variants for each effect type.
                """
            ),
            helptext=(
                """
                This plot shows the effect of variants with respect to
                the mRNA.
                """
            ),
            anchor="snpeff-effects",
            plot=self.effects_plot(),
        )
        self.add_section(
            name="Variants by Functional Class",
            description=(
                """
                The stacked bar plot shows the effect of variants and
                the number of variants for each effect type.
                """
            ),
            helptext=(
                """
                This plot shows the effect of variants on the translation of
                the mRNA as protein. There are three possible cases:

                * **Silent**: The amino acid does not change.
                * **Missense**: The amino acid is different.
                * **Nonsense**: The variant generates a stop codon.
                """
            ),
            anchor="snpeff-functional-class",
            plot=self.effects_function_plot(),
        )
        if len(self.snpeff_qualities) > 0:
            self.add_section(
                name="Variant Qualities",
                description=(
                    """
                    The line plot shows the quantity as function of the
                    variant quality score.
                    """
                ),
                helptext=(
                    """
                    The quality score corresponds to the QUAL column of the
                    VCF file. This score is set by the variant caller.
                    """
                ),
                anchor="snpeff-qualities",
                plot=self.qualities_plot(),
            )

    def parse_snpeff_log(self, f):
        """Go through log file looking for snpeff output"""

        keys = {
            "# Summary table": [
                "Genome",
                "Number_of_variants_before_filter",
                "Number_of_known_variants",
                "Number_of_effects",
                "Genome_total_length",
                "Change_rate",
            ],
            "# Effects by impact": ["HIGH", "LOW", "MODERATE", "MODIFIER"],
            "# Effects by functional class": ["MISSENSE", "NONSENSE", "SILENT", "Missense_Silent_ratio"],
            "# Hom/Het table": ["Het", "Hom", "Missing"],
            "# Ts/Tv summary": ["Transitions", "Transversions", "Ts_Tv_ratio"],
            "# Count by effects": "all",
            "# Count by genomic region": "all",
        }
        parsed_data = {}
        section = None
        version = None
        for line in f["f"]:
            line = line.strip()

            # Parse version
            match = re.search(VERSION_REGEX, line)
            if match:
                version = match.group(1)

            if line[:1] == "#":
                section = line
                self.snpeff_section_totals[section] = dict()
                continue
            s = line.split(",")

            # Quality values / counts
            if section == "# Quality":
                quals = dict()
                if line.startswith("Values"):
                    values = [int(c) for c in line.split(",")[1:]]
                    counts = f["f"].readline()
                    counts = [int(c) for c in counts.split(",")[1:]]
                    c = 0
                    total = sum(counts)
                    for i, v in enumerate(values):
                        if c < (total * 0.995):
                            quals[v] = counts[i]
                            c += counts[i]
                if len(quals) > 0:
                    self.snpeff_qualities[f["s_name"]] = quals

            # Everything else
            elif section in keys:
                if keys[section] == "all" or any([k in s[0].strip() for k in keys[section]]):
                    try:
                        parsed_data[s[0].strip()] = float(s[1].strip())
                    except ValueError:
                        parsed_data[s[0].strip()] = s[1].strip()
                    except IndexError:
                        pass
                    else:
                        # Parsing the number worked - add to totals
                        try:
                            self.snpeff_section_totals[section][s[0].strip()] += parsed_data[s[0].strip()]
                        except KeyError:
                            self.snpeff_section_totals[section][s[0].strip()] = parsed_data[s[0].strip()]
                    if len(s) > 2 and s[2][-1:] == "%":
                        parsed_data[f"{s[0].strip()}_percent"] = float(s[2][:-1])

        if len(parsed_data) > 0:
            if f["s_name"] in self.snpeff_data:
                log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
            self.add_data_source(f)
            self.snpeff_data[f["s_name"]] = parsed_data
            self.add_software_version(version, f["s_name"])

    def general_stats(self):
        """Add key SnpEff stats to the general stats table"""

        headers = {
            "Change_rate": {"title": "Change rate", "scale": "RdYlBu-rev", "min": 0, "format": "{:,.0f}"},
            "Ts_Tv_ratio": {
                "title": "Ts/Tv",
                "description": "Transitions / Transversions ratio",
                "format": "{:,.3f}",
            },
            "Number_of_variants_before_filter": {
                "title": "M Variants",
                "description": "Number of variants before filter (millions)",
                "scale": "PuRd",
                "modify": lambda x: x / 1000000,
                "min": 0,
                "format": "{:,.2f}",
                "suffix": "M",
            },
        }
        self.general_stats_addcols(self.snpeff_data, headers)

    def count_genomic_region_plot(self):
        """Generate the SnpEff Counts by Genomic Region plot"""

        # Sort the keys based on the total counts
        keys = self.snpeff_section_totals["# Count by genomic region"]
        sorted_keys = sorted(keys, reverse=True, key=keys.get)

        # Make nicer label names
        pkeys = dict()
        for k in sorted_keys:
            key = k.replace("_", " ").title().replace("Utr", "UTR")
            if key == "None":
                key = "Unknown"
            pkeys[k] = {"name": key}

        # Config for the plot
        pconfig = {
            "id": "snpeff_variant_effects_region",
            "title": "SnpEff: Counts by Genomic Region",
            "ylab": "# Reads",
            "logswitch": True,
        }

        return bargraph.plot(self.snpeff_data, pkeys, pconfig)

    def effects_plot(self):
        """Generate the SnpEff Counts by Effects plot"""

        # Sort the keys based on the total counts
        keys = self.snpeff_section_totals["# Count by effects"]
        sorted_keys = sorted(keys, reverse=True, key=keys.get)

        # Make nicer label names
        pkeys = dict()
        for k in sorted_keys:
            pkeys[k] = {"name": k.replace("_", " ").title().replace("Utr", "UTR")}

        # Config for the plot
        pconfig = {
            "id": "snpeff_effects",
            "title": "SnpEff: Counts by Effect Types",
            "ylab": "# Reads",
            "logswitch": False,
        }
        return bargraph.plot(self.snpeff_data, pkeys, pconfig)

    def effects_impact_plot(self):
        """Generate the SnpEff Counts by Effects Impact plot"""

        # Put keys in a more logical order
        keys = ["MODIFIER", "LOW", "MODERATE", "HIGH"]

        # Make nicer label names
        pkeys = dict()
        for k in keys:
            pkeys[k] = {"name": k.title()}

        # Config for the plot
        pconfig = {
            "id": "snpeff_variant_effects_impact",
            "title": "SnpEff: Counts by Effects Impact",
            "ylab": "# Reads",
            "logswitch": True,
        }

        return bargraph.plot(self.snpeff_data, pkeys, pconfig)

    def effects_function_plot(self):
        """Generate the SnpEff Counts by Functional Class plot"""

        # Cats to plot in a sensible order
        keys = ["SILENT", "MISSENSE", "NONSENSE"]

        # Make nicer label names
        pkeys = dict()
        for k in keys:
            pkeys[k] = {"name": k.title()}

        # Config for the plot
        pconfig = {
            "id": "snpeff_variant_effects_class",
            "title": "SnpEff: Counts by Functional Class",
            "ylab": "# Reads",
            "logswitch": True,
        }

        return bargraph.plot(self.snpeff_data, pkeys, pconfig)

    def qualities_plot(self):
        """Generate the qualities plot"""

        pconfig = {
            "smooth_points": 200,
            "id": "snpeff_qualities",
            "title": "SnpEff: Qualities",
            "ylab": "Count",
            "xlab": "Values",
            "xDecimals": False,
            "ymin": 0,
            "xmin": 0,
        }

        return linegraph.plot(self.snpeff_qualities, pconfig)



""" MultiQC Submodule to parse output from Qualimap BamQC """


import logging
import os

import math
import re

from multiqc import config
from multiqc.modules.qualimap import parse_numerals
from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(self):
    """Find Qualimap BamQC reports and parse their data"""

    # General stats - genome_results.txt
    self.qualimap_bamqc_genome_results = dict()
    for f in self.find_log_files("qualimap/bamqc/genome_results"):
        parse_genome_results(self, f)

    self.qualimap_bamqc_genome_results = self.ignore_samples(self.qualimap_bamqc_genome_results)
    if len(self.qualimap_bamqc_genome_results) > 0:
        self.write_data_file(self.qualimap_bamqc_genome_results, "multiqc_qualimap_bamqc_genome_results")

    # Coverage - coverage_histogram.txt
    self.qualimap_bamqc_coverage_hist = dict()
    for f in self.find_log_files("qualimap/bamqc/coverage", filehandles=True):
        parse_coverage(self, f)
    self.qualimap_bamqc_coverage_hist = self.ignore_samples(self.qualimap_bamqc_coverage_hist)

    # Insert size - insert_size_histogram.txt
    self.qualimap_bamqc_insert_size_hist = dict()
    for f in self.find_log_files("qualimap/bamqc/insert_size", filehandles=True):
        parse_insert_size(self, f)
    self.qualimap_bamqc_insert_size_hist = self.ignore_samples(self.qualimap_bamqc_insert_size_hist)

    # GC distribution - mapped_reads_gc-content_distribution.txt
    self.qualimap_bamqc_gc_content_dist = dict()
    self.qualimap_bamqc_gc_by_species = dict()  # {'HUMAN': data_dict, 'MOUSE': data_dict}
    for f in self.find_log_files("qualimap/bamqc/gc_dist", filehandles=True):
        parse_gc_dist(self, f)
    self.qualimap_bamqc_gc_content_dist = self.ignore_samples(self.qualimap_bamqc_gc_content_dist)
    self.qualimap_bamqc_gc_by_species = self.ignore_samples(self.qualimap_bamqc_gc_by_species)

    num_parsed = max(
        len(self.qualimap_bamqc_genome_results),
        len(self.qualimap_bamqc_coverage_hist),
        len(self.qualimap_bamqc_insert_size_hist),
        len(self.qualimap_bamqc_gc_content_dist),
    )
    # Go no further if nothing found
    if num_parsed == 0:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    self.add_software_version(None)

    try:
        covs = config.qualimap_config["general_stats_coverage"]
        assert isinstance(covs, list)
        assert len(covs) > 0
        covs = [str(i) for i in covs]
        log.debug(f"Custom Qualimap thresholds: {', '.join([i for i in covs])}")
    except (AttributeError, TypeError, AssertionError):
        covs = [1, 5, 10, 30, 50]
        covs = [str(i) for i in covs]
        log.debug(f"Using default Qualimap thresholds: {', '.join([i for i in covs])}")
    self.covs = covs

    # Make the plots for the report
    report_sections(self)

    # Set up the general stats table
    general_stats_headers(self)

    # Return the number of reports we found
    return num_parsed


def parse_genome_results(self, f):
    """Parse the contents of the Qualimap BamQC genome_results.txt file"""

    int_metrics = {
        "number of reads": "total_reads",
        "number of mapped reads": "mapped_reads",
        "number of mapped reads inside": "regions_mapped_reads",
        "number of mapped bases": "mapped_bases",
        "number of sequenced bases": "sequenced_bases",
        "regions size": "regions_size",
    }
    float_metrics = {
        "mean insert size": "mean_insert_size",
        "median insert size": "median_insert_size",
        "mean mapping quality": "mean_mapping_quality",
        "mean coverageData": "mean_coverage",
    }
    rate_metrics = {  # useful to determine decimal separator
        "general error rate": "general_error_rate",
        "GC percentage": "gc_percentage",
        "duplication rate": "duplication_rate",
        "homopolymer indels": "homopolymer_indels",
    }

    value_regex = re.compile(r"\s+[\d,\.\xa0]+\s+")
    preparsed_d = dict()
    # Keeping track of the section because "number of mapped reads" occurs both
    # under "Globals" and "Globals inside"
    section = None
    for line in f["f"].splitlines():
        if line.startswith(">>>>>>>"):
            section = line[8:]
        elif not section:
            continue
        if "=" in line:
            key, val = line.split("=", 1)
            m = re.search(value_regex, val)
            if m:
                val = m.group(0)
                val = re.sub(r"\xa0", "", val)
            key = key.strip()
            if "Globals inside" in section and key == "number of mapped reads":
                key = "number of mapped reads inside"
            preparsed_d[key] = val.strip()

    # Check we have an input filename
    if "bam file" not in preparsed_d:
        log.debug(f"Couldn't find an input filename in genome_results file {f['fn']}")
        return None
    s_name = self.clean_s_name(preparsed_d["bam file"], f)

    d = parse_numerals(
        preparsed_d,
        float_metrics=float_metrics,
        int_metrics=int_metrics,
        rate_metrics=rate_metrics,
        fpath=os.path.join(f["root"], f["fn"]),
    )

    if "general_error_rate" in d:
        d["general_error_rate"] = d["general_error_rate"] * 100.0
    if "mapped_reads" in d and "total_reads" in d and d["total_reads"] > 0:
        d["percentage_aligned"] = (d["mapped_reads"] / d["total_reads"]) * 100.0
    if "regions_mapped_reads" in d and "mapped_reads" in d and d["mapped_reads"] > 0:
        d["percentage_aligned_on_target"] = (d["regions_mapped_reads"] / d["mapped_reads"]) * 100.0

    # Add to general stats table & calculate a nice % aligned
    for k in [
        "total_reads",
        "mapped_reads",
        "general_error_rate",
        "mean_coverage",
        "regions_size",
        "regions_mapped_reads",
        "percentage_aligned",
        "percentage_aligned_on_target",
    ]:
        if k in d:
            self.general_stats_data[s_name][k] = d[k]

    # Save results
    if s_name in self.qualimap_bamqc_genome_results:
        log.debug(f"Duplicate genome results sample name found! Overwriting: {s_name}")
    self.qualimap_bamqc_genome_results[s_name] = d
    self.add_data_source(f, s_name=s_name, section="genome_results")


def parse_coverage(self, f):
    """Parse the contents of the Qualimap BamQC Coverage Histogram file"""
    # Get the sample name from the parent directory
    # Typical path: <sample name>/raw_data_qualimapReport/coverage_histogram.txt
    s_name = self.get_s_name(f)

    d = dict()
    for line in f["f"]:
        if line.startswith("#"):
            continue
        coverage, count = line.split(None, 1)
        coverage = int(round(float(coverage.replace(",", "."))))
        count = float(count)
        d[coverage] = count

    if len(d) == 0:
        log.debug(f"Couldn't parse contents of coverage histogram file {f['fn']}")
        return None

    # Find median without importing anything to do it for us
    num_counts = sum(d.values())
    cum_counts = 0
    total_cov = 0
    median_coverage = None
    for thiscov, thiscount in d.items():
        cum_counts += thiscount
        total_cov += thiscov * thiscount
        if cum_counts >= num_counts / 2:
            median_coverage = thiscov
            break
    self.general_stats_data[s_name]["median_coverage"] = median_coverage
    # Save results
    if s_name in self.qualimap_bamqc_coverage_hist:
        log.debug(f"Duplicate coverage histogram sample name found! Overwriting: {s_name}")
    self.qualimap_bamqc_coverage_hist[s_name] = d
    self.add_data_source(f, s_name=s_name, section="coverage_histogram")


def parse_insert_size(self, f):
    """Parse the contents of the Qualimap BamQC Insert Size Histogram file"""
    # Get the sample name from the parent directory
    # Typical path: <sample name>/raw_data_qualimapReport/insert_size_histogram.txt
    s_name = self.get_s_name(f)

    d = dict()
    for line in f["f"]:
        if line.startswith("#"):
            continue
        insertsize, count = line.split(None, 1)
        insertsize = int(round(float(insertsize)))
        count = float(count) / 1000000
        if insertsize != 0:
            d[insertsize] = count

    # Find median without importing anything to do it for us
    num_counts = sum(d.values())
    cum_counts = 0
    median_insert_size = None
    for thisins, thiscount in d.items():
        cum_counts += thiscount
        if cum_counts >= num_counts / 2:
            median_insert_size = thisins
            break
    # Add the median insert size to the general stats table
    self.general_stats_data[s_name]["median_insert_size"] = median_insert_size

    # Save results
    if s_name in self.qualimap_bamqc_insert_size_hist:
        log.debug(f"Duplicate insert size histogram sample name found! Overwriting: {s_name}")
    self.qualimap_bamqc_insert_size_hist[s_name] = d
    self.add_data_source(f, s_name=s_name, section="insert_size_histogram")


def parse_gc_dist(self, f):
    """Parse the contents of the Qualimap BamQC Mapped Reads GC content distribution file"""
    # Get the sample name from the parent directory
    # Typical path: <sample name>/raw_data_qualimapReport/mapped_reads_gc-content_distribution.txt
    s_name = self.get_s_name(f)

    d = dict()
    reference_species = None
    reference_d = dict()
    avg_gc = 0
    for line in f["f"]:
        if line.startswith("#"):
            sections = line.strip("\n").split("\t", 3)
            if len(sections) > 2:
                reference_species = sections[2]
            continue
        sections = line.strip("\n").split("\t", 3)
        gc = int(round(float(sections[0])))
        content = float(sections[1])
        avg_gc += gc * content
        d[gc] = content
        if len(sections) > 2:
            reference_content = float(sections[2])
            reference_d[gc] = reference_content

    # Add average GC to the general stats table
    self.general_stats_data[s_name]["avg_gc"] = avg_gc

    # Save results
    if s_name in self.qualimap_bamqc_gc_content_dist:
        log.debug(f"Duplicate Mapped Reads GC content distribution sample name found! Overwriting: {s_name}")
    self.qualimap_bamqc_gc_content_dist[s_name] = d
    if reference_species and reference_species not in self.qualimap_bamqc_gc_by_species:
        self.qualimap_bamqc_gc_by_species[reference_species] = reference_d
    self.add_data_source(f, s_name=s_name, section="mapped_gc_distribution")


coverage_histogram_helptext = """
For a set of DNA or RNA reads mapped to a reference sequence, such as a genome
or transcriptome, the depth of coverage at a given base position is the number
of high-quality reads that map to the reference at that position
(<a href="https://doi.org/10.1038/nrg3642" target="_blank">Sims et al. 2014</a>).

Bases of a reference sequence (y-axis) are groupped by their depth of coverage
(*0&#215;, 1&#215;, &#8230;, N&#215;*) (x-axis). This plot shows
the frequency of coverage depths relative to the reference sequence for each
read dataset, which provides an indirect measure of the level and variation of
coverage depth in the corresponding sequenced sample.

If reads are randomly distributed across the reference sequence, this plot
should resemble a Poisson distribution (<a href="https://doi.org/10.1016/0888-7543(88)90007-9"
target="_blank">Lander & Waterman 1988</a>), with a peak indicating approximate
depth of coverage, and more uniform coverage depth being reflected in a narrower
spread. The optimal level of coverage depth depends on the aims of the
experiment, though it should at minimum be sufficiently high to adequately
address the biological question; greater uniformity of coverage is generally
desirable, because it increases breadth of coverage for a given depth of
coverage, allowing equivalent results to be achieved at a lower sequencing depth
(<a href="https://doi.org/10.1002/gepi.20575" target="_blank">Sampson
et al. 2011</a>; <a href="https://doi.org/10.1038/nrg3642" target="_blank">Sims
et al. 2014</a>). However, it is difficult to achieve uniform coverage
depth in practice, due to biases introduced during sample preparation
(<a href="https://doi.org/10.1016/j.yexcr.2014.01.008" target="_blank">van
Dijk et al. 2014</a>), sequencing (<a href="https://doi.org/10.1186/gb-2013-14-5-r51"
target="_blank">Ross et al. 2013</a>) and read mapping
(<a href="https://doi.org/10.1038/nrg3642" target="_blank">Sims et al. 2014</a>).

This plot may include a small peak for regions of the reference sequence with
zero depth of coverage. Such regions may be absent from the given sample (due
to a deletion or structural rearrangement), present in the sample but not
successfully sequenced (due to bias in sequencing or preparation), or sequenced
but not successfully mapped to the reference (due to the choice of mapping
algorithm, the presence of repeat sequences, or mismatches caused by variants
or sequencing errors). Related factors cause most datasets to contain some
unmapped reads (<a href="https://doi.org/10.1038/nrg3642" target="_blank">Sims
et al. 2014</a>)."""

genome_fraction_helptext = """
For a set of DNA or RNA reads mapped to a reference sequence, such as a genome
or transcriptome, the depth of coverage at a given base position is the number
of high-quality reads that map to the reference at that position, while the
breadth of coverage is the fraction of the reference sequence to which reads
have been mapped with at least a given depth of coverage
(<a href="https://doi.org/10.1038/nrg3642" target="_blank">Sims et al. 2014</a>).

Defining coverage breadth in terms of coverage depth is useful, because
sequencing experiments typically require a specific minimum depth of coverage
over the region of interest (<a href="https://doi.org/10.1038/nrg3642"
target="_blank">Sims et al. 2014</a>), so the extent of the reference sequence
that is amenable to analysis is constrained to lie within regions that have
sufficient depth. With inadequate sequencing breadth, it can be difficult to
distinguish the absence of a biological feature (such as a gene) from a lack
of data (<a href="https://doi.org/10.1101/gr.7050807" target="_blank">Green 2007</a>).

For increasing coverage depths (*1&#215;, 2&#215;, &#8230;, N&#215;*),
coverage breadth is calculated as the percentage of the reference
sequence that is covered by at least that number of reads, then plots
coverage breadth (y-axis) against coverage depth (x-axis). This plot
shows the relationship between sequencing depth and breadth for each read
dataset, which can be used to gauge, for example, the likely effect of a
minimum depth filter on the fraction of a genome available for analysis."""


def report_sections(self):
    """Add results from Qualimap BamQC parsing to the report"""
    # Append to self.sections list

    if len(self.qualimap_bamqc_coverage_hist) > 0:
        # Chew back on histogram to prevent long flat tail
        # (find a sensible max x - lose 1% of longest tail)
        max_x = 0
        total_bases_by_sample = dict()
        for s_name, d in self.qualimap_bamqc_coverage_hist.items():
            total_bases_by_sample[s_name] = sum(d.values())
            cumulative = 0
            for count in sorted(d.keys(), reverse=True):
                cumulative += d[count]
                if cumulative / total_bases_by_sample[s_name] > 0.01:
                    max_x = max(max_x, count)
                    break

        rates_within_threshs = dict()
        for s_name, hist in self.qualimap_bamqc_coverage_hist.items():
            total = total_bases_by_sample[s_name]
            # Make a range of depths that isn't stupidly huge for high coverage expts
            depth_range = list(range(0, max_x + 1, math.ceil(float(max_x) / 400.0) if max_x > 0 else 1))
            # Check that we have our specified coverages in the list
            for c in self.covs:
                if int(c) not in depth_range:
                    depth_range.append(int(c))
            # Calculate the coverage rates for this range of coverages
            rates_within_threshs[s_name] = _calculate_bases_within_thresholds(hist, total, depth_range)
            # Add requested coverage levels to the General Statistics table
            for c in self.covs:
                if int(c) in rates_within_threshs[s_name]:
                    self.general_stats_data[s_name][f"{c}_x_pc"] = rates_within_threshs[s_name][int(c)]
                else:
                    self.general_stats_data[s_name][f"{c}_x_pc"] = 0

        # Section 1 - BamQC Coverage Histogram
        self.add_section(
            name="Coverage histogram",
            anchor="qualimap-coverage-histogram",
            description="Distribution of the number of locations in the reference genome with a given depth of coverage.",
            helptext=coverage_histogram_helptext,
            plot=linegraph.plot(
                self.qualimap_bamqc_coverage_hist,
                {
                    "id": "qualimap_coverage_histogram",
                    "title": "Qualimap BamQC: Coverage histogram",
                    "ylab": "Genome bin counts",
                    "xlab": "Coverage (X)",
                    "ymin": 0,
                    "xmin": 0,
                    "xmax": max_x,
                    "xDecimals": False,
                    "tt_label": "<b>{point.x}X</b>: {point.y}",
                },
            ),
        )
        # Section 2 - BamQC cumulative coverage genome fraction
        self.add_section(
            name="Cumulative genome coverage",
            anchor="qualimap-cumulative-genome-fraction-coverage",
            description="Percentage of the reference genome with at least the given depth of coverage.",
            helptext=genome_fraction_helptext,
            plot=linegraph.plot(
                rates_within_threshs,
                {
                    "id": "qualimap_genome_fraction",
                    "title": "Qualimap BamQC: Genome fraction covered by at least X reads",
                    "ylab": "Fraction of reference (%)",
                    "xlab": "Coverage (X)",
                    "ymax": 100,
                    "ymin": 0,
                    "xmin": 0,
                    "xmax": max_x,
                    "xDecimals": False,
                    "tt_label": "<b>{point.x}X</b>: {point.y:.2f}%",
                },
            ),
        )

    # Section 3 - Insert size histogram
    if len(self.qualimap_bamqc_insert_size_hist) > 0:
        insert_size_helptext = """
        To overcome limitations in the length of DNA or RNA sequencing reads,
        many sequencing instruments can produce two or more shorter reads from
        one longer fragment in which the relative position of reads is
        approximately known, such as paired-end or mate-pair reads
        (<a href="https://doi.org/10.1146/annurev-anchem-062012-092628"
        target="_blank">Mardis 2013</a>). Such techniques can extend the reach
        of sequencing technology, allowing for more accurate placement of reads
        (<a href="https://doi.org/10.1146/annurev-genom-090413-025358"
        target="_blank">Reinert et al. 2015</a>) and better resolution of repeat
        regions (<a href="https://doi.org/10.1146/annurev-genom-090413-025358"
        target="_blank">Reinert et al. 2015</a>), as well as detection of
        structural variation (<a href="https://doi.org/10.1038/nrg2958"
        target="_blank">Alkan et al. 2011</a>) and chimeric transcripts
        (<a href="https://doi.org/10.1073/pnas.0904720106"
        target="_blank">Maher et al. 2009</a>).

        All these methods assume that the approximate size of an insert is known.
        (Insert size can be defined as the length in bases of a sequenced DNA or
        RNA fragment, excluding technical sequences such as adapters, which are
        typically removed before alignment.) This plot allows for that assumption
        to be assessed. With the set of mapped fragments for a given sample, QualiMap
        groups the fragments by insert size, then plots the frequency of mapped
        fragments (y-axis) over a range of insert sizes (x-axis). In an ideal case,
        the distribution of fragment sizes for a sequencing library would culminate
        in a single peak indicating average insert size, with a narrow spread
        indicating highly consistent fragment lengths.

        QualiMap calculates insert sizes as follows: for each fragment in which
        every read mapped successfully to the same reference sequence, it
        extracts the insert size from the `TLEN` field of the leftmost read
        (see the <a href="http://qualimap.bioinfo.cipf.es/doc_html/index.html"
        target="_blank">Qualimap 2 documentation</a>), where the `TLEN` (or
        'observed Template LENgth') field contains 'the number of bases from the
        leftmost mapped base to the rightmost mapped base'
        (<a href="https://samtools.github.io/hts-specs/" target="_blank">SAM
        format specification</a>). Note that because it is defined in terms of
        alignment to a reference sequence, the value of the `TLEN` field may
        differ from the insert size due to factors such as alignment clipping,
        alignment errors, or structural variation or splicing in a gap between
        reads from the same fragment."""
        self.add_section(
            name="Insert size histogram",
            anchor="qualimap-insert-size-histogram",
            description="Distribution of estimated insert sizes of mapped reads.",
            helptext=insert_size_helptext,
            plot=linegraph.plot(
                self.qualimap_bamqc_insert_size_hist,
                {
                    "id": "qualimap_insert_size",
                    "title": "Qualimap BamQC: Insert size histogram",
                    "ylab": "Fraction of reads",
                    "xlab": "Insert Size (bp)",
                    "ymin": 0,
                    "xmin": 0,
                    "tt_label": "<b>{point.x} bp</b>: {point.y}",
                },
            ),
        )

    # Section 4 - GC-content distribution
    if len(self.qualimap_bamqc_gc_content_dist) > 0:
        gc_content_helptext = """
        GC bias is the difference between the guanine-cytosine content
        (GC-content) of a set of sequencing reads and the GC-content of the DNA
        or RNA in the original sample. It is a well-known issue with sequencing
        systems, and may be introduced by PCR amplification, among other factors
        (<a href="https://doi.org/10.1093/nar/gks001" target="_blank">Benjamini
        & Speed 2012</a>; <a href="https://doi.org/10.1186/gb-2013-14-5-r51"
        target="_blank">Ross et al. 2013</a>).

        QualiMap calculates the GC-content of individual mapped reads, then
        groups those reads by their GC-content (*1%, 2%, &#8230;, 100%*), and
        plots the frequency of mapped reads (y-axis) at each level of GC-content
        (x-axis). This plot shows the GC-content distribution of mapped reads
        for each read dataset, which should ideally resemble that of the
        original sample. It can be useful to display the GC-content distribution
        of an appropriate reference sequence for comparison, and QualiMap has an
        option to do this (see the <a href="http://qualimap.bioinfo.cipf.es/doc_html/index.html"
        target="_blank">Qualimap 2 documentation</a>)."""
        extra_series = []
        for i, (species_name, species_data) in enumerate(sorted(self.qualimap_bamqc_gc_by_species.items())):
            extra_series.append(
                {
                    "name": species_name,
                    "data": list(species_data.items()),
                    "dashStyle": "Dash",
                    "lineWidth": 1,
                    "color": ["#000000", "#E89191"][i % 2],
                }
            )
        if len(self.qualimap_bamqc_gc_content_dist) == 1:
            desc = "The solid line represents the distribution of GC content of mapped reads for the sample."
        else:
            desc = "Each solid line represents the distribution of GC content of mapped reads for a given sample."
        lg_config = {
            "id": "qualimap_gc_content",
            "title": "Qualimap BamQC: GC content distribution",
            "ylab": "Fraction of reads",
            "xlab": "GC content (%)",
            "ymin": 0,
            "xmin": 0,
            "xmax": 100,
            "tt_label": "<b>{point.x}%</b>: {point.y:.3f}",
        }
        if len(extra_series) == 1:
            desc += " The dotted line represents a pre-calculated GC distribution for the reference genome."
            lg_config["extra_series"] = extra_series
        elif len(extra_series) > 1:
            desc += " Each dotted line represents a pre-calculated GC distribution for a specific reference genome."
            lg_config["extra_series"] = extra_series

        self.add_section(
            name="GC content distribution",
            anchor="qualimap-gc-distribution",
            description=desc,
            helptext=gc_content_helptext,
            plot=linegraph.plot(self.qualimap_bamqc_gc_content_dist, lg_config),
        )


def general_stats_headers(self):
    try:
        hidecovs = config.qualimap_config["general_stats_coverage_hidden"]
        assert isinstance(hidecovs, list)
        log.debug(f"Hiding Qualimap thresholds: {', '.join([i for i in hidecovs])}")
    except (AttributeError, TypeError, KeyError, AssertionError):
        hidecovs = [1, 5, 10, 50]
    hidecovs = [str(i) for i in hidecovs]

    self.general_stats_headers["avg_gc"] = {
        "title": "% GC",
        "description": "Mean GC content",
        "max": 100,
        "min": 0,
        "suffix": "%",
        "scale": "PuRd",
        "format": "{:,.0f}",
    }
    self.general_stats_headers["median_insert_size"] = {
        "title": "Ins. size",
        "description": "Median insert size",
        "min": 0,
        "scale": "PuOr",
        "format": "{:,.0f}",
    }
    for c in self.covs:
        self.general_stats_headers[f"{c}_x_pc"] = {
            "title": f"≥ {c}X",
            "description": f"Fraction of genome with at least {c}X coverage",
            "max": 100,
            "min": 0,
            "suffix": "%",
            "scale": "RdYlGn",
            "hidden": c in hidecovs,
        }
    self.general_stats_headers["median_coverage"] = {
        "title": "Median cov",
        "description": "Median coverage",
        "min": 0,
        "suffix": "X",
        "scale": "BuPu",
    }
    self.general_stats_headers["mean_coverage"] = {
        "title": "Mean cov",
        "description": "Mean coverage",
        "min": 0,
        "suffix": "X",
        "scale": "BuPu",
    }
    self.general_stats_headers["percentage_aligned_on_target"] = {
        "title": "% On target",
        "description": "% mapped reads on target region",
        "max": 100,
        "min": 0,
        "suffix": "%",
        "scale": "YlGn",
    }
    self.general_stats_headers["regions_size"] = {
        "title": f"{config.read_count_prefix} Region size",
        "description": "Size of target region",
        "suffix": " bp",
        "scale": "PuBuGn",
        "hidden": True,
    }
    self.general_stats_headers["regions_mapped_reads"] = {
        "title": f"{config.read_count_prefix} On target",
        "description": f"Number of mapped reads on target region ({config.read_count_desc})",
        "scale": "RdYlGn",
        "shared_key": "read_count",
        "hidden": True,
    }
    self.general_stats_headers["general_error_rate"] = {
        "title": "Error rate",
        "description": "Alignment error rate. Total edit distance (SAM NM field) over the number of mapped bases",
        "max": 100,
        "min": 0,
        "suffix": "%",
        "scale": "OrRd",
        "format": "{0:.2f}",
        "hidden": True,
    }
    self.general_stats_headers["percentage_aligned"] = {
        "title": "% Aligned",
        "description": "% mapped reads",
        "max": 100,
        "min": 0,
        "suffix": "%",
        "scale": "YlGn",
    }
    self.general_stats_headers["mapped_reads"] = {
        "title": f"{config.read_count_prefix} Aligned",
        "description": f"Number of mapped reads ({config.read_count_desc})",
        "scale": "RdYlGn",
        "shared_key": "read_count",
        "hidden": True,
    }
    self.general_stats_headers["total_reads"] = {
        "title": f"{config.read_count_prefix} Total reads",
        "description": f"Number of reads ({config.read_count_desc})",
        "scale": "Blues",
        "shared_key": "read_count",
        "hidden": True,
    }


def _calculate_bases_within_thresholds(bases_by_depth, total_size, depth_thresholds):
    bases_within_threshs = {depth: 0 for depth in depth_thresholds}
    rates_within_threshs = {depth: None for depth in depth_thresholds}

    dt = sorted(depth_thresholds, reverse=True)
    c = 0
    for depth in sorted(bases_by_depth.keys(), reverse=True):
        while depth < dt[c]:
            c += 1
            bases_within_threshs[dt[c]] = bases_within_threshs[dt[c - 1]]
        if depth >= dt[c]:
            bases_within_threshs[dt[c]] += bases_by_depth[depth]
    while c + 1 < len(dt):
        c += 1
        bases_within_threshs[dt[c]] = total_size
    for t in dt:
        bs = bases_within_threshs[t]
        if total_size > 0:
            rate = 100.0 * bases_within_threshs[t] / total_size
            assert rate <= 100, (
                "Error: rate is > 1: rate = " + str(rate) + ", bases = " + str(bs) + ", size = " + str(total_size)
            )
            rates_within_threshs[t] = rate
    return rates_within_threshs



""" MultiQC module to parse output from Prokka """


import logging

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, table

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Prokka",
            anchor="prokka",
            href="http://www.vicbioinformatics.com/software.prokka.shtml",
            info="is a software tool for the rapid annotation of prokaryotic genomes.",
            doi="10.1093/bioinformatics/btu153",
        )

        # Parse logs
        self.prokka = dict()
        for f in self.find_log_files("prokka", filehandles=True):
            self.parse_prokka(f)

        # Filter to strip out ignored sample names
        self.prokka = self.ignore_samples(self.prokka)

        if len(self.prokka) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.prokka)} logs")

        self.write_data_file(self.prokka, "multiqc_prokka")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Add most important Prokka annotation stats to the general table
        headers = {
            "organism": {
                "title": "Organism",
                "description": "Organism",
            },
            "contigs": {
                "title": "Contigs",
                "description": "Number of contigs",
                "min": 0,
            },
            "bases": {
                "title": "Bases",
                "description": "Number of bases",
                "min": 0,
                "format": "{:i}%",
                "hidden": True,
            },
            "CDS": {
                "title": "CDS",
                "description": "Number of CDS",
                "min": 0,
                "format": "{:i}%",
            },
        }
        self.general_stats_addcols(self.prokka, headers)

        # User can set configuration attributes, 'prokka_table', and
        # 'prokka_barplot', to specify whether to include a table or a barplot, or both.
        # Default is to make a plot only.
        if getattr(config, "prokka_table", False):
            self.add_section(plot=self.prokka_table())
        if getattr(config, "prokka_barplot", True):
            descr_plot = "This barplot shows the distribution of different types of features found in each contig."
            helptext = """
            `Prokka` can detect different features:

            - CDS
            - rRNA
            - tmRNA
            - tRNA
            - miscRNA
            - signal peptides
            - CRISPR arrays

            This barplot shows you the distribution of these different types of features found in each contig.
            """
            self.add_section(plot=self.prokka_barplot(), helptext=helptext, description=descr_plot)

    def parse_prokka(self, f):
        """Parse prokka txt summary files.

        Prokka summary files are difficult to identify as there are practically
        no distinct prokka identifiers in the filenames or file contents. This
        parser makes an attempt using the first three lines, expected to contain
        organism, contigs, and bases statistics.
        """

        s_name = None

        # Look at the first three lines, they are always the same
        first_line = f["f"].readline()
        contigs_line = f["f"].readline()
        bases_line = f["f"].readline()
        # If any of these fail, it's probably not a prokka summary file
        if not all(
            (first_line.startswith("organism:"), contigs_line.startswith("contigs:"), bases_line.startswith("bases:"))
        ):
            return

        # Get organism and sample name from the first line
        # Assumes organism name only consists of two words,
        # i.e. 'Genusname speciesname', and that the remaining
        # text on the organism line is the sample name.
        try:
            organism = " ".join(first_line.strip().split(":", 1)[1].split()[:2])
            s_name = self.clean_s_name(" ".join(first_line.split()[3:]), f)
        except KeyError:
            organism = first_line.strip().split(":", 1)[1]
            s_name = f["s_name"]
        # Don't try to guess sample name if requested in the config
        if getattr(config, "prokka_fn_snames", False):
            s_name = f["s_name"]

        if s_name in self.prokka:
            log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
        self.prokka[s_name] = dict()
        self.prokka[s_name]["organism"] = organism
        self.prokka[s_name]["contigs"] = int(contigs_line.split(":")[1])
        self.prokka[s_name]["bases"] = int(bases_line.split(":")[1])

        # Get additional info from remaining lines
        for line in f["f"]:
            description, value = line.split(":")
            try:
                self.prokka[s_name][description] = int(value)
            except ValueError:
                log.warning("Unable to parse line: '%s'", line)

        self.add_data_source(f, s_name)

    def prokka_table(self):
        """Make basic table of the annotation stats"""

        # Specify the order of the different possible categories
        headers = {
            "organism": {
                "title": "Organism",
                "description": "Organism name",
            },
            "contigs": {
                "title": "# contigs",
                "description": "Number of contigs in assembly",
                "format": "{:i}",
            },
            "bases": {
                "title": "# bases",
                "description": "Number of nucleotide bases in assembly",
                "format": "{:i}",
            },
            "CDS": {
                "title": "# CDS",
                "description": "Number of annotated CDS",
                "format": "{:i}",
            },
            "rRNA": {
                "title": "# rRNA",
                "description": "Number of annotated rRNA",
                "format": "{:i}",
            },
            "tRNA": {
                "title": "# tRNA",
                "description": "Number of annotated tRNA",
                "format": "{:i}",
            },
            "tmRNA": {
                "title": "# tmRNA",
                "description": "Number of annotated tmRNA",
                "format": "{:i}",
            },
            "misc_RNA": {
                "title": "# misc RNA",
                "description": "Number of annotated misc. RNA",
                "format": "{:i}",
            },
            "sig_peptide": {
                "title": "# sig_peptide",
                "description": "Number of annotated sig_peptide",
                "format": "{:i}",
            },
            "repeat_region": {
                "title": "# CRISPR arrays",
                "description": "Number of annotated CRSIPR arrays",
                "format": "{:i}",
            },
        }
        table_config = {
            "namespace": "prokka",
            "min": 0,
        }

        return table.plot(self.prokka, headers, table_config)

    def prokka_barplot(self):
        """Make a basic plot of the annotation stats"""

        # Specify the order of the different categories
        keys = {
            "CDS": {"name": "CDS"},
            "rRNA": {"name": "rRNA"},
            "tRNA": {"name": "tRNA"},
            "tmRNA": {"name": "tmRNA"},
            "misc_RNA": {"name": "misc RNA"},
            "sig_peptide": {"name": "Signal peptides"},
            "repeat_region": {"name": "CRISPR array"},
        }

        plot_config = {
            "id": "prokka_plot",
            "title": "Prokka: Feature Types",
            "ylab": "# Counts",
            "cpswitch_counts_label": "Features",
        }

        return bargraph.plot(self.prokka, keys, plot_config)



""" MultiQC submodule to parse output from Picard RnaSeqMetrics """

import logging

from multiqc.modules.picard import util
from multiqc.plots import bargraph, linegraph

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(module):
    """Find Picard RnaSeqMetrics reports and parse their data"""

    data_by_sample = dict()
    histogram_by_sample = dict()

    # Go through logs and find Metrics
    for f in module.find_log_files("picard/rnaseqmetrics", filehandles=True):
        # Sample name from input file name by default.
        s_name = f["s_name"]
        in_hist = False

        for line in f["f"]:
            maybe_s_name = util.extract_sample_name(
                module,
                line,
                f,
                picard_tool="RnaSeqMetrics",
            )
            if maybe_s_name:
                s_name = maybe_s_name

            if s_name is None:
                continue

            # Catch the histogram values
            if in_hist:
                try:
                    sections = line.split("\t")
                    pos = int(sections[0])
                    coverage = float(sections[1])
                    histogram_by_sample[s_name][pos] = coverage
                except ValueError:
                    # Reset in case we have more in this log file
                    s_name = None
                    in_hist = False

            if util.is_line_right_before_table(line, picard_class="RnaSeqMetrics"):
                keys = f["f"].readline().strip("\n").split("\t")
                vals = f["f"].readline().strip("\n").split("\t")
                if len(vals) != len(keys):
                    continue

                if s_name in data_by_sample:
                    log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")

                module.add_data_source(f, s_name, section="RnaSeqMetrics")
                data_by_sample[s_name] = dict()
                histogram_by_sample[s_name] = dict()

                for k, v in zip(keys, vals):
                    if not v:
                        v = "NA"
                    else:
                        try:
                            v = float(v)
                        except ValueError:
                            pass
                        else:
                            # Multiply percentages by 100
                            if k.startswith("PCT_"):
                                v = v * 100.0
                    data_by_sample[s_name][k] = v
                # Calculate some extra numbers
                if "PF_BASES" in keys and "PF_ALIGNED_BASES" in keys:
                    data_by_sample[s_name]["PF_NOT_ALIGNED_BASES"] = (
                        data_by_sample[s_name]["PF_BASES"] - data_by_sample[s_name]["PF_ALIGNED_BASES"]
                    )

            elif line.startswith("## HISTOGRAM"):
                keys = f["f"].readline().strip("\n").split("\t")
                assert len(keys) >= 2, (keys, f)
                in_hist = True
                histogram_by_sample[s_name] = dict()

    # Filter to strip out ignored sample names
    data_by_sample = module.ignore_samples(data_by_sample)
    histogram_by_sample = module.ignore_samples(histogram_by_sample)
    if len(data_by_sample) == 0:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    module.add_software_version(None)

    # Write parsed data to a file
    module.write_data_file(data_by_sample, "multiqc_picard_RnaSeqMetrics")

    # Add to general stats table
    headers = {
        "PCT_RIBOSOMAL_BASES": {
            "title": "rRNA",
            "description": "Percent of aligned bases overlapping ribosomal RNA regions",
            "max": 100,
            "min": 0,
            "suffix": "%",
            "scale": "Reds",
        },
        "PCT_MRNA_BASES": {
            "title": "mRNA",
            "description": "Percent of aligned bases overlapping UTRs and coding regions of mRNA transcripts",
            "max": 100,
            "min": 0,
            "suffix": "%",
            "scale": "Greens",
        },
    }
    module.general_stats_addcols(data_by_sample, headers, namespace="RnaSeqMetrics")

    # Bar plot of bases assignment
    bg_cats = {
        "CODING_BASES": {"name": "Coding"},
        "UTR_BASES": {"name": "UTR"},
        "INTRONIC_BASES": {"name": "Intronic"},
        "INTERGENIC_BASES": {"name": "Intergenic"},
        "RIBOSOMAL_BASES": {"name": "Ribosomal"},
        "PF_NOT_ALIGNED_BASES": {"name": "PF not aligned"},
    }

    # Warn user if any samples are missing 'RIBOSOMAL_BASES' data; ie picard was run without an rRNA interval file.
    warn_rrna = ""
    rrna_missing = []
    for s_name, metrics in data_by_sample.items():
        if metrics["RIBOSOMAL_BASES"] == "NA":
            rrna_missing.append(s_name)
    if rrna_missing:
        if len(rrna_missing) < 5:
            missing_samples = f"for samples <code>{'</code>, <code>'.join(rrna_missing)}</code>"
        else:
            missing_samples = f"<strong>{len(rrna_missing)} samples</strong>"
        warn_rrna = f"""
        <div class="alert alert-warning">
          <span class="glyphicon glyphicon-warning-sign"></span>
          Picard was run without an rRNA annotation file {missing_samples}, therefore the ribosomal assignment is not available. To correct, rerun with the <code>RIBOSOMAL_INTERVALS</code> parameter, as documented <a href="https://broadinstitute.github.io/picard/command-line-overview.html#CollectRnaSeqMetrics" target="_blank">here</a>.
        </div>
        """

    pconfig = {
        "id": "picard_rnaseqmetrics_assignment_plot",
        "title": "Picard: RnaSeqMetrics Base Assignments",
        "ylab": "Number of bases",
    }
    module.add_section(
        name="RnaSeqMetrics Assignment",
        anchor="picard-rna-assignment",
        description="Number of bases in primary alignments that align to regions in the reference genome." + warn_rrna,
        plot=bargraph.plot(data_by_sample, bg_cats, pconfig),
    )

    # Bar plot of strand mapping
    bg_cats = dict()
    bg_cats["CORRECT_STRAND_READS"] = {"name": "Correct"}
    bg_cats["INCORRECT_STRAND_READS"] = {"name": "Incorrect", "color": "#8e123c"}

    pdata = dict()
    for s_name, d in data_by_sample.items():
        if d["CORRECT_STRAND_READS"] > 0 and d["INCORRECT_STRAND_READS"] > 0:
            pdata[s_name] = d
    if len(pdata) > 0:
        pconfig = {
            "id": "picard_rnaseqmetrics_strand_plot",
            "title": "Picard: RnaSeqMetrics Strand Mapping",
            "ylab": "Number of reads",
            "hide_zero_cats": False,
        }
        module.add_section(
            name="RnaSeqMetrics Strand Mapping",
            anchor="picard-rna-strand",
            description="Number of aligned reads that map to the correct strand.",
            plot=bargraph.plot(data_by_sample, bg_cats, pconfig),
        )

    # Section with histogram plot
    if len(histogram_by_sample) > 0:
        # Plot the data and add section
        pconfig = {
            "smooth_points": 500,
            "smooth_points_sumcounts": [True, False],
            "id": "picard_rna_coverage",
            "title": "Picard: Normalized Gene Coverage",
            "ylab": "Coverage",
            "xlab": "Percent through gene",
            "xDecimals": False,
            "tt_label": "<b>{point.x}%</b>: {point.y:.0f}",
            "ymin": 0,
        }
        module.add_section(
            name="Gene Coverage",
            anchor="picard-rna-coverage",
            plot=linegraph.plot(histogram_by_sample, pconfig),
        )

    # Return the number of detected samples to the parent module
    return len(data_by_sample)



""" MultiQC Submodule to parse output from Qualimap RNASeq """


import logging
import os
import re

from multiqc import config
from multiqc.modules.qualimap import parse_numerals
from multiqc.plots import bargraph, linegraph

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(self):
    """Find Qualimap RNASeq reports and parse their data"""
    self.qualimap_rnaseq_genome_results = dict()

    int_metrics = {
        "read pairs aligned": "reads_aligned",
        "reads aligned": "reads_aligned",
        "total alignments": "total_alignments",
        "non-unique alignments": "non_unique_alignments",
        "aligned to genes": "reads_aligned_genes",
        "ambiguous alignments": "ambiguous_alignments",
        "not aligned": "not_aligned",
        "exonic": "reads_aligned_exonic",
        "intronic": "reads_aligned_intronic",
        "intergenic": "reads_aligned_intergenic",
        "overlapping exon": "reads_aligned_overlapping_exon",
    }
    rate_metrics = {
        "5' bias": "5_bias",
        "3' bias": "3_bias",
        "SSP estimation (fwd/rev)": "ssp_estimation",
        "5'-3' bias": "5_3_bias",
    }

    value_regex = re.compile(r"\s+[\d,\.\xa0]+\s+")
    for f in self.find_log_files("qualimap/rnaseq/rnaseq_results"):
        preparsed_d = dict()
        for line in f["f"].splitlines():
            if "=" in line:
                key, val = line.split("=", 1)
                m = re.search(value_regex, val)
                if m:
                    val = m.group(0)
                    val = re.sub(r"\xa0", "", val)
                key = key.strip()
                preparsed_d[key] = val.strip()

        # Check we have an input filename
        if "bam file" not in preparsed_d:
            log.debug(f"Couldn't find an input filename in genome_results file {f['fn']}")
            return None
        s_name = self.clean_s_name(preparsed_d["bam file"], f)

        d = parse_numerals(
            preparsed_d,
            float_metrics={},
            int_metrics=int_metrics,
            rate_metrics=rate_metrics,
            fpath=os.path.join(f["root"], f["fn"]),
        )

        # Add to general stats table
        for k in ["5_3_bias", "reads_aligned"]:
            try:
                self.general_stats_data[s_name][k] = d[k]
            except KeyError:
                pass

        # Save results
        if s_name in self.qualimap_rnaseq_genome_results:
            log.debug(f"Duplicate genome results sample name found! Overwriting: {s_name}")
        self.qualimap_rnaseq_genome_results[s_name] = d
        self.add_data_source(f, s_name=s_name, section="rna_genome_results")

    # Coverage profile
    self.qualimap_rnaseq_cov_hist = dict()
    for f in self.find_log_files("qualimap/rnaseq/coverage", filehandles=True):
        s_name = self.get_s_name(f)
        d = dict()
        for line in f["f"]:
            if line.startswith("#"):
                continue
            coverage, count = line.split(None, 1)
            coverage = int(round(float(coverage.replace(",", "."))))
            count = float(count)
            d[coverage] = count

        if len(d) == 0:
            log.debug(f"Couldn't parse contents of coverage histogram file {f['fn']}")
            return None

        # Save results
        if s_name in self.qualimap_rnaseq_cov_hist:
            log.debug(f"Duplicate coverage histogram sample name found! Overwriting: {s_name}")
        self.qualimap_rnaseq_cov_hist[s_name] = d
        self.add_data_source(f, s_name=s_name, section="rna_coverage_histogram")

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    self.add_software_version(None)

    # Filter to strip out ignored sample names
    self.qualimap_rnaseq_genome_results = self.ignore_samples(self.qualimap_rnaseq_genome_results)
    self.qualimap_rnaseq_cov_hist = self.ignore_samples(self.qualimap_rnaseq_cov_hist)

    # Plots
    # Genomic Origin Bar Graph
    # NB: Ignore 'Overlapping Exon' in report - these make the numbers add up to > 100%
    if len(self.qualimap_rnaseq_genome_results) > 0:
        # Check that we have anything to plot
        if (
            sum(
                entry[key]
                for entry in self.qualimap_rnaseq_genome_results.values()
                for key in ["reads_aligned_exonic", "reads_aligned_intronic", "reads_aligned_intronic"]
            )
            > 0
        ):
            # Write data to file
            self.write_data_file(self.qualimap_rnaseq_genome_results, "qualimap_rnaseq_genome_results")

            gorigin_cats = {
                "reads_aligned_exonic": {"name": "Exonic"},
                "reads_aligned_intronic": {"name": "Intronic"},
                "reads_aligned_intergenic": {"name": "Intergenic"},
            }
            gorigin_pconfig = {
                "id": "qualimap_genomic_origin",
                "title": "Qualimap RNAseq: Genomic Origin",
                "ylab": "Number of reads",
                "cpswitch_c_active": False,
            }
            genomic_origin_helptext = """
            There are currently three main approaches to map reads to transcripts in an
            RNA-seq experiment: mapping reads to a reference genome to identify expressed
            transcripts that are annotated (and discover those that are unknown), mapping
            reads to a reference transcriptome, and <i>de novo</i> assembly of transcript
            sequences (<a href="https://doi.org/10.1186/s13059-016-0881-8"
            target="_blank">Conesa et al. 2016</a>).

            For RNA-seq QC analysis, QualiMap can be used to assess alignments produced by
            the first of these approaches. For input, it requires a GTF annotation file
            along with a reference genome, which can be used to reconstruct the exon
            structure of known transcripts. This allows mapped reads to be grouped by
            whether they originate in an exonic region (for QualiMap, this may include
            5&#8242; and 3&#8242; UTR regions as well as protein-coding exons), an intron,
            or an intergenic region (see the <a href="http://qualimap.bioinfo.cipf.es/doc_html/index.html"
            target="_blank">Qualimap 2 documentation</a>).

            The inferred genomic origins of RNA-seq reads are presented here as a bar graph
            showing either the number or percentage of mapped reads in each read dataset
            that have been assigned to each type of genomic region. This graph can be used
            to assess the proportion of useful reads in an RNA-seq experiment. That
            proportion can be reduced by the presence of intron sequences, especially if
            depletion of ribosomal RNA was used during sample preparation (<a href="https://doi.org/10.1038/nrg3642"
            target="_blank">Sims et al. 2014</a>). It can also be reduced by off-target
            transcripts, which are detected in greater numbers at the sequencing depths
            needed to detect poorly-expressed transcripts (<a href="https://doi.org/10.1101/gr.124321.111"
            target="_blank">Tarazona et al. 2011</a>)."""
            self.add_section(
                name="Genomic origin of reads",
                anchor="qualimap-reads-genomic-origin",
                description="Classification of mapped reads as originating in exonic, intronic or intergenic regions. These can be displayed as either the number or percentage of mapped reads.",
                helptext=genomic_origin_helptext,
                plot=bargraph.plot(self.qualimap_rnaseq_genome_results, gorigin_cats, gorigin_pconfig),
            )
        else:
            log.warning("Found zero aligned reads. Skipping 'Genomic origin of reads' plot.")

    if len(self.qualimap_rnaseq_cov_hist) > 0:
        # Write data to file
        self.write_data_file(self.qualimap_rnaseq_cov_hist, "qualimap_rnaseq_cov_hist")

        # Make a normalised percentage version of the coverage data
        self.qualimap_rnaseq_cov_hist_percent = dict()
        for s_name in self.qualimap_rnaseq_cov_hist:
            self.qualimap_rnaseq_cov_hist_percent[s_name] = dict()
            total = sum(self.qualimap_rnaseq_cov_hist[s_name].values())
            if total == 0:
                for k, v in self.qualimap_rnaseq_cov_hist[s_name].items():
                    self.qualimap_rnaseq_cov_hist_percent[s_name][k] = 0.0
            else:
                for k, v in self.qualimap_rnaseq_cov_hist[s_name].items():
                    self.qualimap_rnaseq_cov_hist_percent[s_name][k] = (v / total) * 100.0

        coverage_profile_helptext = """
        There are currently three main approaches to map reads to transcripts in an
        RNA-seq experiment: mapping reads to a reference genome to identify expressed
        transcripts that are annotated (and discover those that are unknown), mapping
        reads to a reference transcriptome, and <i>de novo</i> assembly of transcript
        sequences (<a href="https://doi.org/10.1186/s13059-016-0881-8"
        target="_blank">Conesa et al. 2016</a>).

        For RNA-seq QC analysis, QualiMap can be used to assess alignments produced by
        the first of these approaches. For input, it requires a GTF annotation file
        along with a reference genome, which can be used to reconstruct the exon
        structure of known transcripts. QualiMap uses this information to calculate the
        depth of coverage along the length of each annotated transcript. For a set of
        reads mapped to a transcript, the depth of coverage at a given base position is
        the number of high-quality reads that map to the transcript at that position
        (<a href="https://doi.org/10.1038/nrg3642" target="_blank">Sims et al. 2014</a>).

        QualiMap calculates coverage depth at every base position of each annotated
        transcript. To enable meaningful comparison between transcripts, base positions
        are rescaled to relative positions expressed as percentage distance along each
        transcript (*0%, 1%, &#8230;, 99%*). For the set of transcripts with at least
        one mapped read, QualiMap plots the _cumulative mapped-read depth_ (y-axis) at
        each relative transcript position (x-axis). This plot shows the gene coverage
        profile across all mapped transcripts for each read dataset. It provides a
        visual way to assess positional biases, such as an accumulation of mapped reads
        at the 3&#8242; end of transcripts, which may indicate poor RNA quality in the
        original sample (<a href="https://doi.org/10.1186/s13059-016-0881-8"
        target="_blank">Conesa et al. 2016</a>).

        The _Normalised_ plot is calculated by MultiQC to enable comparison of samples
        with varying sequencing depth. The _cumulative mapped-read depth_ at each
        position across the averaged transcript position are divided by the total for
        that sample across the entire averaged transcript.
        """
        pconfig = {
            "id": "qualimap_gene_coverage_profile",
            "title": "Qualimap RNAseq: Coverage Profile Along Genes (total)",
            "ylab": "Cumulative mapped-read depth",
            "xlab": "Transcript Position (%)",
            "ymin": 0,
            "xmin": 0,
            "xmax": 100,
            "tt_label": "<b>{point.x}%</b>: {point.y:.2f}",
            "data_labels": [
                {"name": "Counts", "ylab": "Cumulative mapped-read depth"},
                {"name": "Normalised", "ylab": "Percentage total cumulative mapped-read depth"},
            ],
        }
        self.add_section(
            name="Gene Coverage Profile",
            anchor="qualimap-genome-fraction-coverage",
            description="Mean distribution of coverage depth across the length of all mapped transcripts.",
            helptext=coverage_profile_helptext,
            plot=linegraph.plot([self.qualimap_rnaseq_cov_hist, self.qualimap_rnaseq_cov_hist_percent], pconfig),
        )

    #### General Stats
    self.general_stats_headers["5_3_bias"] = {
        "title": "5'-3' bias",
        "format": "{:,.2f}",
    }
    self.general_stats_headers["reads_aligned"] = {
        "title": f"{config.read_count_prefix} Aligned",
        "description": f"Reads Aligned ({config.read_count_desc})",
        "min": 0,
        "scale": "RdBu",
        "shared_key": "read_count",
        "modify": lambda x: x * config.read_count_multiplier,
    }

    # Return the number of reports we found
    return len(self.qualimap_rnaseq_genome_results.keys())
